"""ratelimit.py

created 31-dec-2020 by richb@instantlinux.net
"""

from base64 import b64encode
from flask import g, request
import hashlib
import logging
from redis.exceptions import ConnectionError
import time

from . import state
from .grants import Grants
from .service_config import ServiceConfig


class RateLimit(object):
    """Rate Limiting

    Args:
      enable (bool): enable limits [default: config.RATELIMIT_ENABLE]
      interval (int): seconds to count limit [config.RATELIMIT_INTERVAL]
    """
    def __init__(self, enable=False, interval=None):
        self.config = ServiceConfig().config
        self.enable = enable or self.config.RATELIMIT_ENABLE
        self.interval = interval or self.config.RATELIMIT_INTERVAL
        self.redis = state.redis_conn

    def call(self, limit=None, service='*', uid=None):
        """Apply the granted rate limit for uid to a given service
        Two redis keys are used, for even and odd intervals as measured
        modulo the Unix epoch time. The current interval's key is
        incremented if we haven't reached the limit yet, then we delete
        that key at first call seen next interval and increment the other
        key. The keys expire within interval-1 seconds to keep the cache clear
        after the user stops sending calls.

        Args:
          limit (int): limit for interval [default: from Grants]
          service (str): name of a service
          uid (str): a user ID

        Returns:
          bool: true if limit exceeded
        """
        if not self.enable:
            return False
        uid = uid or self._get_request_uid()
        if not uid:
            # No limit on anonymous requests, TODO make this smarter to
            #  protect against brute-force DDOS / security attacks
            return False
        limit = limit or Grants().get('ratelimit', uid=uid)
        curr = round(time.time()) % (self.interval * 2) // self.interval
        key = 'rate:%s:%s:%d' % (service, uid, curr)
        altkey = 'rate:%s:%s:%d' % (service, uid, 0 if curr else 1)
        try:
            calls = self.redis.get(key)
        except ConnectionError as ex:
            logging.error(dict(action='ratelimit.call', message=str(ex)))
            return False
        if limit and not calls or int(calls.decode()) < limit:
            try:
                pipe = self.redis.pipeline()
                pipe.incr(key).expire(
                    key, self.interval - 1).delete(altkey).execute()
            except Exception as ex:
                logging.error(dict(action='ratelimit.call', message=str(ex)))
        else:
            logging.info(dict(action='ratelimit.call', uid=uid,
                              service=service, message='limit exceeded'))
            return True
        return False

    def reset(self, service='*', uid=None):
        """Clear the current entry for a service

        Args:
          service (str): name of a service
          uid (str): a user ID
        """
        self.redis.delete('rate:%s:%s:0' % (service, uid))
        self.redis.delete('rate:%s:%s:1' % (service, uid))

    def _get_request_uid(self):
        """Examine headers to find uid
        """
        header_auth = self.config.HEADER_AUTH_APIKEY
        if header_auth in request.headers:
            try:
                prefix, secret = request.headers.get(header_auth).split('.')
            except ValueError:
                return None
            secret = b64encode(
                hashlib.sha1(secret.encode()).digest())[:8].decode('utf8')
            # TODO confirm whether this drops first use of apikey
            return g.session.get(None, secret, key_id=prefix, arg='sub')
        elif request.authorization:
            return request.authorization.username



"""format.py

created 15-aug-2020 by docker@instantlinux.net
"""

from babel.support import format_datetime, Translations
from datetime import datetime
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText
from email.utils import make_msgid, formatdate
import jinja2
import logging
from sqlalchemy.orm.exc import NoResultFound

from ..exceptions import APIcrudFormatError
from ..service_config import ServiceConfig

jinja2Env = []


class MessageFormat(object):
    """MIME and SMS formatting
    """

    def __init__(self, db_session=None, account_id=None, smtp=None,
                 settings=None):
        # TODO this could be subclassed from Messaging - but causes
        # circular import exception
        self.config = ServiceConfig().config
        self.db_session = db_session
        self.models = ServiceConfig().models
        self.settings = settings
        self.siteurl = self.settings.get.url

    def email(self, template, frm, sender_email, to, attachments=[], **kwargs):
        """Format an email message with mime attachment, in the user's
        profile-determined locale

        Args:
            template (str): jinja2 template name
            frm (Contact): contact of sender
            sender_email (str): destination address
            to (Contact): contact of recipient
            attachments (list): additional MIMEBase / MimeText objects
            **kwargs: other key=value pairs for jinja2 substitution
        Returns:
            obj: mime object (use .as_string() to convert to message)
        """
        params = dict(
            sender=frm.owner.name, **kwargs)
        Profile = ServiceConfig().models.Profile
        try:
            tz = self.db_session.query(Profile).filter(
                Profile.uid == to.uid, Profile.item == 'tz').one().value
        except NoResultFound:
            tz = self.settings.get.tz
        try:
            lang = self.db_session.query(Profile).filter(
                Profile.uid == to.uid, Profile.item == 'lang').one().value
        except NoResultFound:
            lang = self.settings.get.lang
        if 'starts' in kwargs:
            params['starts_formatted'] = format_datetime(
                datetime.strptime(kwargs['starts'], '%Y-%m-%dT%H:%M:%S'),
                'EEEE, d MMM Y H:mm a', locale=lang, tzinfo=tz)
        mime = MIMEMultipart('alternative')
        mime['From'] = '%s <%s>' % (frm.owner.name, sender_email)
        if frm.info != sender_email:
            mime['Reply-To'] = '%s <%s>' % (frm.owner.name, frm.info)
        mime['To'] = '%s <%s>' % (to.owner.name, to.info)
        mime['Date'] = formatdate()
        mime['Subject'] = self._render_template(template, lang,
                                                selector='subject', **kwargs)
        mime['Message-ID'] = make_msgid()
        mime.attach(MIMEText(
            self._render_template(template, lang, selector='email', **params) +
            self._render_template('footer', lang, selector='email', **params),
            'plain'))
        mime.attach(MIMEText(
            self._render_template(template, lang, selector='html', **params) +
            self._render_template('footer', lang, selector='html', **params),
            'html'))
        for item in attachments:
            mime.attach(item)
        return mime

    def sms(self, template, frm, to, **kwargs):
        """Format an SMS message

        Args:
            template (str): jinja2 template name
            frm (Contact): contact of sender
            to (Contact): contact of recipient
            **kwargs: other key=value pairs for jinja2 substitution
        Returns:
            str: Formatted string content
        """
        Profile = ServiceConfig().models.Profile
        try:
            lang = self.db_session.query(Profile).filter(
                Profile.uid == to.uid, Profile.item == 'lang').one().value
        except NoResultFound:
            lang = self.settings.get.lang
        return self._render_template(template, lang, selector='sms',
                                     sender=frm.owner.name, **kwargs)

    def _render_template(self, template_name, locale, **kwargs):
        """Scan the configured template_folders path looking for the
        template, apply translations from locale, and process
        with jinja2.

        Args:
            template_name (str): filename of template (without .j2)
            locale (str): language/locale string
            kwargs: key/value pairs for template expansion
        Returns:
            str: Rendered content
        Raises:
            APIcrudFormatError: template not found
        """
        global jinja2Env

        config = ServiceConfig().config
        trans = []
        for dir in config.BABEL_TRANSLATION_DIRECTORIES.split(';'):
            trans.append(Translations.load(dir, locale))
        if not jinja2Env:
            for folder in config.TEMPLATE_FOLDERS:
                jinja2Env.append(jinja2.Environment(
                    loader=jinja2.FileSystemLoader(searchpath=folder),
                    extensions=['jinja2.ext.i18n', 'jinja2.ext.autoescape'],
                    autoescape=jinja2.select_autoescape(['html'])
                ))

        kwargs.update(dict(siteurl=self.siteurl, appname=self.config.APPNAME))
        for env in jinja2Env:
            try:
                for x in trans:
                    env.install_gettext_translations(x)
                return env.get_template(template_name + '.j2').render(**kwargs)
            except jinja2.exceptions.TemplateNotFound:
                continue
        logging.error(dict(message='template not found',
                           template=template_name))
        raise APIcrudFormatError('template=%s not found' % template_name)



"""auth controller

created 6-apr-2019 by richb@instantlinux.net
"""

from flask import g, request
from flask_babel import _
import logging

from apicrud import SessionAuth, state
from apicrud.auth import AuthTOTP, OAuth2


class AuthController(object):

    def __init__(self):
        self.resource = 'auth'
        if self.resource not in state.controllers:
            state.controllers[self.resource] = self

    @staticmethod
    def login(body):
        """Login a new session

        Args:
          body (dict): specify method, or username and password
        Returns:
          dict:
            Fields include jwt_token (contains uid / account ID),
            ID of entry in settings database, and a sub-dictionary
            with mapping of endpoints registered to microservices
        """
        return SessionAuth().account_login(
            body.get('username'), body.get('password'),
            method=body.get('method'), otp=body.get('otp'),
            nonce=body.get('nonce'))

    def logout():
        """Logout

        Returns:
          tuple: dict with message and http status code 200
        """
        creds = request.authorization
        if creds:
            g.session.delete(creds.username, creds.password)
            logging.info('action=logout id=%s' % creds.username)
        return dict(message=_(u'logged out')), 200

    @staticmethod
    def auth_callback(method, code, state):
        return OAuth2().callback(method, code=code, state=state)

    @staticmethod
    def auth_params():
        """Get authorization info"""
        return SessionAuth().auth_params()

    @staticmethod
    def totp_generate():
        """Generate a new TOTP token"""
        return AuthTOTP().generate()

    @staticmethod
    def totp_register(body):
        """Register TOTP for user"""
        return AuthTOTP().register(body)

    @staticmethod
    def find():
        """Find auth methods - for the frontend's Login screen"""
        return SessionAuth().methods()



"""grant controller

created 27-may-2019 by richb@instantlinux.net
"""

from apicrud import BasicCRUD, Grants


class GrantController(BasicCRUD):
    def __init__(self):
        super().__init__(resource='grant')

    @staticmethod
    def create(body):
        if not body.get('expires'):
            body.pop('expires', None)
        ret = super(GrantController, GrantController).create(body)
        Grants().uncache(body.get('uid'))
        return ret

    @staticmethod
    def update(id, body):
        """If the id has a hybrid uid:grant syntax, invoke create
        instead; otherwise it's a standard update

        Args:
            id (str): Database or hybrid grant ID
            body (dict): resource fields as defined by openapi.yaml schema
        """
        Grants().uncache(body.get('uid'))
        if not body.get('expires'):
            body['expires'] = None
        if ':' in id:
            body['uid'] = id.split(':')[0]
            body.pop('id', None)
            return super(GrantController, GrantController).create(body)
        return super(GrantController, GrantController).update(id, body)

    @staticmethod
    def get(id):
        """Get one grant

        Args:
            id (str): Database or hybrid grant ID
        """
        return Grants().crud_get(
            super(GrantController, GrantController).get(id), id)

    @staticmethod
    def find(**kwargs):
        """Find multiple grants

        Args:
            kwargs: as defined in openapi.yaml
        """
        return Grants().find(super(GrantController, GrantController).find(
            **kwargs), **kwargs)



"""metric controller

created 27-may-2019 by richb@instantlinux.net
"""

from apicrud import BasicCRUD, Metrics


class MetricController(BasicCRUD):
    def __init__(self):
        self.resource = 'metric'

    @staticmethod
    def find(**kwargs):
        """Find multiple metrics

        Args:
            kwargs: as defined in openapi.yaml
        """
        return Metrics().find(**kwargs)

    @staticmethod
    def collect(**kwargs):
        """Collect metrics in prometheus-compatible format

        Args:
            kwargs: as defined in openapi.yaml
        """
        return Metrics().collect(**kwargs)



"""Open-core API model definitions

created 26-mar-2019 by richb@instantlinux.net

license: lgpl-2.1
"""

# coding: utf-8

# from geoalchemy2 import Geometry
from sqlalchemy import BOOLEAN, Column, Enum, ForeignKey, INTEGER, \
     LargeBinary, String, TEXT, TIMESTAMP, Unicode, UniqueConstraint
from sqlalchemy import func
from sqlalchemy.ext.hybrid import hybrid_property
from sqlalchemy.orm import relationship, backref
from sqlalchemy_utils.types.encrypted.encrypted_type import AesEngine, \
    StringEncryptedType

from apicrud.const import Constants
from .base import get_aes_secret, AsDictMixin, Base
core_schema = 'apicrud'
schema_pre = core_schema + '.' if core_schema else ''

# TODO fix foreign-key dependency-cycle settings / credentials / storage
#  see https://docs.sqlalchemy.org/en/14/core
#   /constraints.html#creating-dropping-foreign-key-constraints-via-alter


class AlembicVersionApicrud(Base):
    __tablename__ = 'alembic_version_apicrud'
    version_num = Column(String(32), primary_key=True, nullable=False)


class Account(AsDictMixin, Base):
    __tablename__ = 'accounts'
    __table_args__ = (
        UniqueConstraint(u'id', u'uid', name='uniq_account_user'),
        dict(schema=core_schema),
    )
    __rest_exclude__ = ('backup_codes', 'password', 'totp_secret')
    __rest_hybrid__ = ('totp',)

    id = Column(String(16), primary_key=True, unique=True)
    name = Column(String(32), nullable=False, unique=True)
    uid = Column(ForeignKey(u'%speople.id' % schema_pre, ondelete='CASCADE'),
                 nullable=False, unique=True)
    password = Column(StringEncryptedType(Unicode, get_aes_secret, AesEngine,
                                          'pkcs5', length=128), nullable=False)
    password_must_change = Column(BOOLEAN, nullable=False,
                                  server_default="False")
    totp_secret = Column(StringEncryptedType(Unicode, get_aes_secret,
                                             AesEngine, 'pkcs5', length=64))
    backup_codes = Column(LargeBinary(256))
    is_admin = Column(BOOLEAN, nullable=False, server_default="False")
    settings_id = Column(ForeignKey(u'%ssettings.id' % schema_pre),
                         nullable=False)
    last_login = Column(TIMESTAMP)
    created = Column(TIMESTAMP, nullable=False, server_default=func.now())
    modified = Column(TIMESTAMP)
    status = Column(Enum(u'active', u'disabled'), nullable=False)

    owner = relationship('Person', foreign_keys=[uid], backref=backref(
        'account_uid', cascade='all, delete-orphan'))
    settings = relationship('Settings')

    @hybrid_property
    def totp(self):
        return True if self.totp_secret else False


class APIkey(AsDictMixin, Base):
    __tablename__ = 'apikeys'
    __table_args__ = (
        UniqueConstraint(u'uid', u'name', name='uniq_apikey_owner'),
        dict(schema=core_schema),
    )
    __rest_exclude__ = ('hashvalue',)
    __rest_related__ = ('scopes',)

    id = Column(String(16), primary_key=True, unique=True)
    name = Column(String(32), nullable=False)
    uid = Column(ForeignKey(u'%speople.id' % schema_pre, ondelete='CASCADE'),
                 nullable=False)
    prefix = Column(String(8), nullable=False, unique=True)
    hashvalue = Column(StringEncryptedType(Unicode, get_aes_secret, AesEngine,
                                           'pkcs5', length=96), nullable=False)
    expires = Column(TIMESTAMP)
    last_used = Column(TIMESTAMP)
    created = Column(TIMESTAMP, nullable=False, server_default=func.now())
    modified = Column(TIMESTAMP)
    status = Column(Enum(u'active', u'disabled'), nullable=False,
                    server_default=u'active')

    scopes = relationship('Scope', secondary='%sapikeyscopes' % schema_pre,
                          backref=backref('scope'), order_by='Scope.name')
    owner = relationship('Person', foreign_keys=[uid], backref=backref(
        'apikey_uid', cascade='all, delete-orphan'))


class APIkeyScope(Base):
    __tablename__ = 'apikeyscopes'
    __table_args__ = (
        UniqueConstraint(u'apikey_id', u'scope_id', name='uniq_apikey_scope'),
        dict(schema=core_schema),
    )

    apikey_id = Column(
        ForeignKey(u'%sapikeys.id' % schema_pre, ondelete='CASCADE'),
        primary_key=True, nullable=False)
    scope_id = Column(
        ForeignKey(u'%sscopes.id' % schema_pre, ondelete='CASCADE'),
        primary_key=True, nullable=False, index=True)
    created = Column(TIMESTAMP, nullable=False, server_default=func.now())

    apikey = relationship('APIkey', foreign_keys=[apikey_id], backref=backref(
        'apikeyscopes', cascade='all, delete-orphan'))
    scope = relationship('Scope', backref=backref(
        'apikeyscopes', cascade='all, delete-orphan'))


class Category(AsDictMixin, Base):
    __tablename__ = 'categories'
    __table_args__ = (
        UniqueConstraint(u'name', u'uid', name='uniq_category_owner'),
        dict(schema=core_schema),
    )

    id = Column(String(16), primary_key=True, unique=True)
    name = Column(String(64), nullable=False)
    uid = Column(ForeignKey(u'%speople.id' % schema_pre), nullable=False)
    created = Column(TIMESTAMP, nullable=False, server_default=func.now())
    modified = Column(TIMESTAMP)
    status = Column(Enum('active', u'disabled'), nullable=False,
                    server_default=u'active')

    owner = relationship('Person', foreign_keys=[uid], backref=backref(
        'category_uid', cascade='all, delete-orphan'))


class Contact(AsDictMixin, Base):
    __tablename__ = 'contacts'
    __table_args__ = (
        UniqueConstraint(u'info', u'type', name='uniq_info_type'),
        dict(schema=core_schema),
    )

    id = Column(String(16), primary_key=True, unique=True)
    label = Column(String(16), nullable=False, server_default=u'home')
    type = Column(String(12), nullable=False, server_default=u'email')
    carrier = Column(String(16))
    info = Column(String(255))
    muted = Column(BOOLEAN, nullable=False, server_default="False")
    rank = Column(INTEGER, nullable=False, server_default="1")
    uid = Column(ForeignKey(u'%speople.id' % schema_pre, ondelete='CASCADE'),
                 nullable=False, index=True)
    privacy = Column(String(8), nullable=False, server_default=u'member')
    created = Column(TIMESTAMP, nullable=False, server_default=func.now())
    modified = Column(TIMESTAMP)
    status = Column(Enum(u'active', u'unconfirmed', u'disabled'),
                    nullable=False, server_default=u'unconfirmed')

    owner = relationship('Person', foreign_keys=[uid], backref=backref(
        'contacts', cascade='all, delete-orphan'))


class Credential(AsDictMixin, Base):
    __tablename__ = 'credentials'
    __table_args__ = (
        UniqueConstraint(u'name', u'uid', name='uniq_name_uid'),
        dict(schema=core_schema),
    )
    __rest_exclude__ = ('secret', 'otherdata')

    id = Column(String(16), primary_key=True, unique=True)
    name = Column(String(64), nullable=False)
    vendor = Column(String(32), nullable=False)
    # placeholder field TODO make clear how this will be used
    #  probably to delegate to Secrets Manager / KMS
    type = Column(String(16))
    url = Column(String(length=64))
    key = Column(String(length=128))
    secret = Column(StringEncryptedType(Unicode, get_aes_secret, AesEngine,
                                        'pkcs5', length=128))
    otherdata = Column(StringEncryptedType(Unicode, get_aes_secret, AesEngine,
                                           'pkcs5', length=1024))
    expires = Column(TIMESTAMP)
    settings_id = Column(ForeignKey(u'%ssettings.id' % schema_pre),
                         nullable=False)
    uid = Column(ForeignKey(u'%speople.id' % schema_pre, ondelete='CASCADE'),
                 nullable=False)
    created = Column(TIMESTAMP, nullable=False, server_default=func.now())
    modified = Column(TIMESTAMP)
    status = Column(Enum(u'active', u'disabled'), nullable=False,
                    server_default=u'active')

    settings = relationship('Settings', foreign_keys=[settings_id])
    owner = relationship('Person', foreign_keys=[uid], backref=backref(
        'credential_uid', cascade='all, delete-orphan'))


class Grant(AsDictMixin, Base):
    __tablename__ = 'grants'
    __table_args__ = (
        UniqueConstraint(u'name', u'uid', name='uniq_grant_user'),
        dict(schema=core_schema),
    )

    id = Column(String(16), primary_key=True, unique=True)
    name = Column(String(24), nullable=False)
    value = Column(String(64), nullable=False)
    uid = Column(ForeignKey(u'%speople.id' % schema_pre), nullable=False)
    expires = Column(TIMESTAMP)
    created = Column(TIMESTAMP, nullable=False, server_default=func.now())
    modified = Column(TIMESTAMP)
    status = Column(Enum('active', u'disabled'), nullable=False,
                    server_default=u'active')

    owner = relationship('Person', foreign_keys=[uid], backref=backref(
        'grant_uid', cascade='all, delete-orphan'))


class List(AsDictMixin, Base):
    __tablename__ = 'lists'
    __table_args__ = (
        UniqueConstraint(u'name', u'uid', name='uniq_list_owner'),
        dict(schema=core_schema),
    )
    __rest_related__ = ('members',)

    id = Column(String(16), primary_key=True, unique=True)
    name = Column(String(64), nullable=False)
    description = Column(TEXT())
    privacy = Column(String(8), nullable=False, server_default=u'secret')
    category_id = Column(ForeignKey(u'%scategories.id' % schema_pre),
                         nullable=False)
    uid = Column(ForeignKey(u'%speople.id' % schema_pre, ondelete='CASCADE'),
                 nullable=False)
    created = Column(TIMESTAMP, nullable=False, server_default=func.now())
    modified = Column(TIMESTAMP)
    status = Column(Enum('active', u'disabled'), nullable=False)

    members = relationship('Person', secondary='listmembers',
                           backref=backref('people'), order_by='Person.name')
    # messages = relationship('Message', secondary='listmessages',
    #                         backref=backref('lists'),
    #                         order_by='Message.created')
    category = relationship('Category')
    owner = relationship('Person', foreign_keys=[uid], backref=backref(
        'list_uid', cascade='all, delete-orphan'))


# see https://docs.sqlalchemy.org/en/13/orm/extensions/associationproxy.html
class ListMember(Base):
    __tablename__ = 'listmembers'
    __table_args__ = (
        UniqueConstraint(u'list_id', u'uid', name='uniq_listmember'),
    )

    uid = Column(ForeignKey(u'%speople.id' % schema_pre, ondelete='CASCADE'),
                 primary_key=True, nullable=False)
    list_id = Column(
        ForeignKey(u'%slists.id' % schema_pre, ondelete='CASCADE'),
        primary_key=True, nullable=False, index=True)
    authorization = Column(String(8), nullable=False,
                           server_default=u'member')
    created = Column(TIMESTAMP, nullable=False, server_default=func.now())

    person = relationship('Person', foreign_keys=[uid], backref=backref(
        'listmembers', cascade='all, delete-orphan'))
    list = relationship('List', backref=backref(
        'listmembers', cascade='all, delete-orphan'))


class Location(AsDictMixin, Base):
    __tablename__ = 'locations'
    __table_args__ = (
        dict(schema=core_schema),
    )

    id = Column(String(16), primary_key=True, unique=True)
    name = Column(String(255))
    address = Column(String(255))
    city = Column(String(64), nullable=False)
    state = Column(String(16))
    postalcode = Column(String(12))
    country = Column(String(2), nullable=False,
                     server_default=Constants.DEFAULT_COUNTRY)
    # MariaDB is not supported by geoalchemy2, alas
    # geo = Column(Geometry(geometry_type='POINT'))
    # Coordinates are stored as fixed-precision (divide by 1e7)
    geolat = Column(INTEGER)
    geolong = Column(INTEGER)
    neighborhood = Column(String(32))
    privacy = Column(String(8), nullable=False, server_default=u'public')
    uid = Column(ForeignKey(u'%speople.id' % schema_pre, ondelete='CASCADE'),
                 nullable=False)
    category_id = Column(ForeignKey(u'%scategories.id' % schema_pre),
                         nullable=False)
    created = Column(TIMESTAMP, nullable=False, server_default=func.now())
    modified = Column(TIMESTAMP)
    status = Column(Enum('active', u'disabled'), nullable=False)

    category = relationship('Category')
    owner = relationship('Person', foreign_keys=[uid], backref=backref(
        'location_uid', cascade='all, delete-orphan'))


class Person(AsDictMixin, Base):
    __tablename__ = 'people'
    __table_args__ = (
        dict(schema=core_schema),
    )
    __rest_related__ = ('lists',)

    id = Column(String(16), primary_key=True, unique=True)
    name = Column(String(64), nullable=False)
    identity = Column(String(64), nullable=False, unique=True)
    referrer_id = Column(ForeignKey(u'%speople.id' % schema_pre))
    privacy = Column(String(8), nullable=False, server_default=u'public')
    created = Column(TIMESTAMP, nullable=False, server_default=func.now())
    modified = Column(TIMESTAMP)
    status = Column(Enum('active', u'disabled', u'suspended'), nullable=False)

    lists = relationship('List', secondary='listmembers',
                         backref=backref('lists'), lazy='dynamic')
    profileitems = relationship('Profile', backref=backref(
        '%sprofileitems' % schema_pre))


class Profile(AsDictMixin, Base):
    __tablename__ = 'profileitems'
    __table_args__ = (
        UniqueConstraint(u'uid', u'item', name='uniq_itemuid'),
        dict(schema=core_schema),
    )

    id = Column(String(16), primary_key=True, unique=True)
    uid = Column(ForeignKey(u'%speople.id' % schema_pre, ondelete='CASCADE'),
                 nullable=False)
    item = Column(String(32), nullable=False)
    value = Column(String(96))
    location_id = Column(ForeignKey(u'%slocations.id' % schema_pre))
    tz_id = Column(ForeignKey(u'%stime_zone_name.id' % schema_pre))
    privacy = Column(String(8), nullable=False, server_default=u'public')
    created = Column(TIMESTAMP, nullable=False, server_default=func.now())
    modified = Column(TIMESTAMP)
    status = Column(Enum('active', u'disabled'), nullable=False)

    location = relationship('Location')
    owner = relationship('Person', foreign_keys=[uid], backref=backref(
        'profile', cascade='all, delete-orphan'))
    tz = relationship('Tz')


class Scope(AsDictMixin, Base):
    __tablename__ = 'scopes'
    __table_args__ = (
        UniqueConstraint(u'name', u'settings_id', name='uniq_scope_name'),
        dict(schema=core_schema),
    )

    id = Column(String(16), primary_key=True, unique=True)
    name = Column(String(32), nullable=False)
    settings_id = Column(ForeignKey(u'%ssettings.id' % schema_pre),
                         nullable=False)
    created = Column(TIMESTAMP, nullable=False, server_default=func.now())
    modified = Column(TIMESTAMP)
    status = Column(Enum('active', u'disabled'), nullable=False,
                    server_default=u'active')


class Settings(AsDictMixin, Base):
    __tablename__ = 'settings'
    __table_args__ = (
        dict(schema=core_schema),
    )

    id = Column(String(16), primary_key=True, unique=True)
    name = Column(String(32), nullable=False)
    privacy = Column(String(8), nullable=False, server_default=u'public')
    smtp_port = Column(INTEGER, nullable=False, server_default='25')
    smtp_smarthost = Column(String(255))
    smtp_credential_id = Column(ForeignKey(
        u'%scredentials.id' % schema_pre, use_alter=True, name='settings_fk1'))
    country = Column(String(2), nullable=False,
                     server_default=Constants.DEFAULT_COUNTRY)
    # TODO storage is for media, need to refactor Settings as a many-to-many
    # like ProfileItems
    default_storage_id = Column(ForeignKey(u'%sstorageitems.id' % schema_pre))
    lang = Column(String(6), nullable=False,
                  server_default=Constants.DEFAULT_LANG)
    tz_id = Column(ForeignKey(u'%stime_zone_name.id' % schema_pre),
                   nullable=False, server_default='598')
    url = Column(String(255))
    window_title = Column(String(127),
                          server_default=Constants.DEFAULT_WINDOW_TITLE)
    default_cat_id = Column(ForeignKey(u'%scategories.id' % schema_pre),
                            nullable=False)
    administrator_id = Column(ForeignKey(u'%speople.id' % schema_pre),
                              nullable=False)
    default_hostlist_id = Column(ForeignKey(u'%slists.id' % schema_pre))
    created = Column(TIMESTAMP, nullable=False, server_default=func.now())
    modified = Column(TIMESTAMP)

    smtp_credential = relationship('Credential',
                                   foreign_keys=[smtp_credential_id])
    default_category = relationship('Category', foreign_keys=[default_cat_id])
    administrator = relationship('Person')
    default_hostlist = relationship('List',
                                    foreign_keys=[default_hostlist_id])
    tz = relationship('Tz')


class Storage(AsDictMixin, Base):
    __tablename__ = 'storageitems'
    __table_args__ = (
        UniqueConstraint(u'name', u'uid', name='uniq_storage_user'),
        dict(schema=core_schema),
    )

    id = Column(String(16), primary_key=True, unique=True)
    name = Column(String(32), nullable=False)
    prefix = Column(String(128))
    bucket = Column(String(64), nullable=False)
    region = Column(String(16),
                    server_default=Constants.DEFAULT_AWS_REGION)
    cdn_uri = Column(String(64))
    identifier = Column(String(64))
    privacy = Column(String(8), nullable=False, server_default=u'public')
    credentials_id = Column(ForeignKey(u'%scredentials.id' % schema_pre))
    uid = Column(ForeignKey(u'%speople.id' % schema_pre, ondelete='CASCADE'),
                 nullable=False)
    created = Column(TIMESTAMP, nullable=False, server_default=func.now())
    modified = Column(TIMESTAMP)
    status = Column(Enum('active', u'disabled'), nullable=False,
                    server_default="active")

    credentials = relationship('Credential')
    owner = relationship('Person', foreign_keys=[uid], backref=backref(
        'storage_uid', cascade='all, delete-orphan'))


class Tz(AsDictMixin, Base):
    __tablename__ = 'time_zone_name'
    __table_args__ = (
        dict(schema=core_schema),
    )

    id = Column(INTEGER, primary_key=True, unique=True)
    name = Column(String(32), nullable=False, unique=True)
    status = Column(Enum('active', u'disabled'), nullable=False,
                    server_default="active")



from flask import g
import logging
from sqlalchemy.orm.exc import NoResultFound

from apicrud import AccountSettings, BasicCRUD, state


class SettingsController(BasicCRUD):
    def __init__(self):
        super().__init__(resource='settings')

    def update(id, body):
        retval = super(SettingsController, SettingsController).update(
            id, body)
        # Clear cached value for global admin
        if retval[1] == 200:
            account_id = None
            try:
                account_id = g.db.query(state.models.Account).filter_by(
                    name='admin').one().id
            except NoResultFound:
                logging.warning('action=update message="Missing admin"')
            AccountSettings(account_id).uncache()
        return retval



"""trashcan.py

created 24-may-2021 by richb@instantlinux.net
"""

from connexion import NoContent
from flask import g, request
from flask_babel import _
import json
import logging
from sqlalchemy import asc, desc
from sqlalchemy.orm.exc import NoResultFound

from .const import Constants
from .database import db_abort
from . import AccessControl, BasicCRUD, state, utils


class Trashcan(BasicCRUD):
    """Deleted-items management

    Read/Update/Delete/find controller operations.

    This class provides permission-based, paginated access to database
    resources that have been marked inactive.
    """
    def __init__(self):
        self.models = state.models
        self.resource = 'trashcan'
        state.controllers[self.resource] = self

    @staticmethod
    def get(id):
        """Controller for GET endpoint. This method evaluates
        privacy settings against the user's permissions, looks up
        category, owner and geocode values, and fetches a single
        record.

        Args:
          id (str): composite ID of the desired resource
        Returns:
          tuple:
            first element is a dict with the object or error
            message, second element is response code (200 on success)
        """
        self = state.controllers[request.url_rule.rule.split('/')[3]]
        resource, record_id = id.split('-', maxsplit=1)
        model = getattr(self.models, resource.capitalize())
        acc = AccessControl(model=model)
        try:
            query = g.db.query(model).filter_by(id=record_id)
            record = query.one()
        except NoResultFound as ex:
            msg = _(u'not found') if 'No row was found' in str(ex) else str(ex)
            return dict(id=id, message=msg), 404

        retval = record.as_dict()
        if hasattr(model, 'uid') and hasattr(model, 'owner'):
            retval['owner'] = record.owner.name
        if 'modified' in retval and not retval['modified']:
            del(retval['modified'])
        eid = None if hasattr(model, 'event_id') else (
            acc.auth_ids[acc.primary_resource] or [None])[0]
        retval['rbac'] = ''.join(sorted(list(
            acc.rbac_permissions(query=query, membership=acc.primary_resource,
                                 id=eid) - set('c'))))
        if 'r' in retval['rbac']:
            logging.info(dict(
                action='get', resource=resource, id=id,
                ident=acc.identity, duration=utils.req_duration()))
            return retval, 200
        else:
            return dict(message=_(u'access denied'), id=id), 403

    @staticmethod
    def update(id, body):
        """Controller for PUT endpoint. This method looks for an existing
        record, evaluates user's permissions, and updates the row in
        the back-end database.

        Args:
          body (dict): fields to be updated
        Returns:
          dict:
            first element is a dict with the id, second element is
            response code (200 on success)
        """
        resource, record_id = id.split('-', maxsplit=1)
        self = state.controllers[request.url_rule.rule.split('/')[3]]
        model = getattr(self.models, resource.capitalize())
        logmsg = dict(action='update', resource=resource, id=record_id,
                      ident=AccessControl().identity)
        try:
            query = g.db.query(model).filter_by(id=record_id)
            if not AccessControl(
                    model=model).with_permission('d', query=query):
                return dict(message=_(u'access denied'), id=id), 403
            current = query.one().as_dict()
            query.update(dict(status=body.get('status')))
            g.db.commit()
        except Exception as ex:
            return db_abort(str(ex), rollback=True, **logmsg)

        updated = {key: body[key] for key, val in body.items()
                   if key in current and current[key] != val and
                   key not in ('name', 'modified')}
        logging.info(dict(name=body.get('name'), duration=utils.req_duration(),
                          **logmsg, **updated))
        return dict(id=id, message=_(u'updated')), 200

    @staticmethod
    def delete(ids):
        """Controller for DELETE endpoints. This method looks for existing
        records, evaluates user's permissions, and updates or removes
        rows in the back-end database.

        Args:
          ids (list of str): composite record IDs to be flagged for removal
        Returns:
          tuple:
            first element is a dict with the id, second element is
            response code (200 on success)
        """

        # TODO - update auth if model could affect any session's auth
        self = state.controllers[request.url_rule.rule.split('/')[3]]
        logmsg = dict(action='delete', resource='trashcan',
                      account_id=AccessControl().account_id,
                      ident=AccessControl().identity)
        errors = 0
        count = 0
        for id in ids:
            resource, record_id = id.split('-', maxsplit=1)
            if resource == 'apikey':
                model = self.models.APIkey
            else:
                model = getattr(self.models, resource.capitalize())
            try:
                query = g.db.query(model).filter_by(id=record_id)
                if not AccessControl(
                        model=model).with_permission(
                            'd', query=query):
                    return dict(message=_(u'access denied'), id=id), 403
                if query.delete():
                    logging.info(dict(id=id, **logmsg))
                else:
                    logging.info(dict(id=id, msg='query failed', **logmsg))
                    errors += 1
            except Exception as ex:
                return db_abort(str(ex), **logmsg)
            count += 1
        try:
            g.db.commit()
        except Exception as ex:
            return db_abort(str(ex), rollback=True, **logmsg)
        logging.info(dict(count=count, ids=ids, duration=utils.req_duration(),
                          **logmsg))
        return NoContent, 404 if errors else 204

    @staticmethod
    def find(**kwargs):
        """Find records which match query parameters passed from
        connexion by name, in a dictionary that also includes user
        and token info

        Due to limitation of react-admin, pagination is provided for
        both offset/limit and cursor/limit. The offset approach requires
        unfortunate hacks and is not recommended: use cursor_next.

        Args:
          cursor_next (str): pagination token to fetch subsequent records
          filter (dict): field/value pairs to query (simple queries
              only, with string or list matching; or * for any)
          limit (int): max records to fetch
          offset (int): old-style pagination starting offset
          sort (str): <field>[:{asc|desc}]
          status (str): value is added to filter
        Returns:
          dict: items (list), count(int), cursor_next (str)
        """

        self = state.controllers[request.url_rule.rule.split('/')[3]]
        acc = AccessControl()
        conditions = {item: value for item, value in kwargs.items()
                      if item in ('status')}
        logmsg = dict(action='find', ident=acc.identity, resource='trashcan',
                      **conditions)
        start_resource = '0'
        if 'cursor_next' in kwargs:
            cursor_next = self._fromb64(kwargs['cursor_next'])
            start_resource = cursor_next.split(':')[1]
            offset = int(cursor_next.split(':')[2])
        elif 'offset' in kwargs:
            offset = int(kwargs['offset'])
        else:
            offset = 0
        sort = kwargs.get('sort', 'name')
        if ':' in sort:
            sort, dir = sort.split(':')
            sortdir = desc if dir == 'desc' else asc
        else:
            sortdir = asc
        try:
            filter = json.loads(kwargs.get('filter', '{}'))
        except json.decoder.JSONDecodeError as ex:
            msg = _(u'invalid filter specified') + '=%s' % str(ex)
            logging.error(dict(message=msg, **logmsg))
            return dict(message=msg), 405
        if 'admin' not in acc.auth:
            if 'uid' in filter and filter['uid'] != acc.uid:
                msg = _(u'access denied')
                logging.info(dict(message=msg, **logmsg))
                return dict(message=msg), 403
            filter['uid'] = acc.uid
        if 'resource' in filter:
            resources = [filter.get('resource')]
            filter.pop('resource')
        else:
            if 'offset' in kwargs:
                return dict(message=_(
                    u'use cursor pagination or specify resource in filter')), \
                    405
            resources = [
                key for key in state.controllers.keys()
                if key not in ('auth', 'scope', 'settings', 'trashcan', 'tz')
                and key >= start_resource]
        if 'id' in filter and not filter.get('id'):
            # allow get:foo?filter{id:null,key=val,etc}
            del(filter['id'])
        limit = logmsg['limit'] = int(kwargs.get(
            'limit', Constants.PER_PAGE_DEFAULT))
        retval = dict(items=[], count=0)
        for resource in resources:
            if resource == 'apikey':
                model = self.models.APIkey
            else:
                model = getattr(self.models, resource.capitalize())
            query = g.db.query(model)
            acc = AccessControl(model=model)
            xfilter = filter.copy()
            for key in filter.keys():
                if key == 'uid' and resource == 'person':
                    xfilter['referrer_id'] = xfilter.pop('uid')
                    continue
                if not hasattr(model, key):
                    msg = _(u'invalid filter key')
                    logging.info(dict(message=msg, **logmsg))
                    return dict(message=msg, resource=resource), 405
                if xfilter[key] == '*':
                    xfilter.pop(key)
                elif (xfilter[key] and type(xfilter[key]) is str and '%'
                      in xfilter[key]):
                    query = query.filter(getattr(model, key).like(
                        xfilter[key]))
                    xfilter.pop(key)
                elif type(xfilter[key]) is list:
                    items = xfilter.pop(key)
                    if len(items) == 1:
                        xfilter[key] = items[0]
                    else:
                        query = query.filter(getattr(model, key).in_(items))
            try:
                query = query.filter_by(**xfilter)
                if hasattr(model, sort):
                    query = query.order_by(sortdir(getattr(model, sort)))
            except Exception as ex:
                return db_abort(str(ex), **logmsg)
            query = query.filter(model.status == 'disabled')
            query = acc.with_filter(query)
            try:
                results = query.slice(offset, offset + limit + 1).all()
            except Exception as ex:
                return db_abort(str(ex), **logmsg)
            for result in results[:limit - retval['count'] + 1]:
                if retval['count'] >= limit:
                    retval['cursor_next'] = self._tob64(
                        'cursor:%s:%d' % (resource, offset))
                    logging.info(dict(
                        offset=offset, duration=utils.req_duration(),
                        **logmsg))
                    return retval, 200
                data = result.as_dict()
                # Compose a limited amount of descriptive metadata
                record = {
                    key: data[key] for key in (
                        'created', 'modified', 'uid', 'status')
                    if hasattr(model, key)}
                record['id'] = '%s-%s' % (resource, data['id'])
                record['resource'] = resource
                record['name'] = ''
                for col in ('name', 'info', 'item', 'subject'):
                    if hasattr(model, col):
                        record['name'] = data[col]
                        break
                # TODO check for delete permission
                record['rbac'] = 'dru'
                retval['items'].append(record)
                retval['count'] += 1
                offset += 1
            if retval['items']:
                offset = 0
        logging.info(dict(
            offset=offset, duration=utils.req_duration(), **logmsg))
        return retval, 200



"""account_settings.py

created 13-may-2019 by richb@instantlinux.net
"""

from cachetools import TTLCache
from collections import namedtuple
import logging
from sqlalchemy.orm.exc import NoResultFound
from sqlalchemy_utils.types.encrypted.padding import InvalidPaddingError

from .service_config import ServiceConfig


class AccountSettings(object):
    """Access class for account settings, with cache - converts db
    record to object attributes

    Each account is associated with an entry in the Settings model; this
    class provides access to these key-value pairs as read-only
    attributes. Because these functions are called frequently, the db
    entry is loaded into memory for a configurable expiration period
    (config.REDIS_TTL).

    Args:
      account_id (str): ID in database of a user's account
      db_session (obj): a session connected to database
      uid (str): User ID
    """
    _singleton = None
    _cache = None
    _admin_id = None

    def __init__(self, account_id, db_session=None, uid=None):
        """Cache per-account settings and convert to attributes"""

        def _convert(attrs):
            return namedtuple('GenericDict', attrs.keys())(**attrs)

        config = ServiceConfig().config
        models = self.models = ServiceConfig().models
        if account_id not in self._cache:
            try:
                if account_id:
                    account = db_session.query(models.Account).filter_by(
                        id=account_id).one()
                    uid = account.uid
                else:
                    account = db_session.query(models.Account).filter_by(
                        uid=uid).one()
                    account_id = account.id
                record = account.settings
            except NoResultFound:
                if self._admin_id:
                    # Already cached
                    return self.__init__(self._admin_id, db_session=db_session)
                try:
                    record = db_session.query(models.Settings).filter_by(
                        name='global').one()
                    account = db_session.query(models.Account).filter_by(
                        uid=record.administrator_id).one()
                except Exception as ex:
                    logging.error(dict(action='AccountSettings',
                                       error=str(ex)))
                    return None
                uid = record.administrator_id
                self._admin_id = account_id = account.id
            except InvalidPaddingError as ex:
                logging.error(dict(action='AccountSettings', error=str(ex)))
                return None
            try:
                category_id = db_session.query(models.Category).filter_by(
                    uid=uid).filter_by(name='default').one().id
            except NoResultFound:
                category_id = record.default_cat_id
            self._cache[account_id] = dict(
                uid=uid,
                settings=dict(
                    record.as_dict(), **dict(
                        category_id=category_id,
                        sender_name=record.administrator.name,
                        sender_email=record.administrator.identity,
                        tz=record.tz.name)))
            try:
                senders = db_session.query(models.List).filter_by(
                    name=config.APPROVED_SENDERS,
                    uid=record.administrator_id).one()
                self._cache[account_id]['settings']['approved_senders'] = (
                    [member.identity for member in senders.members])
            except NoResultFound:
                self._cache[account_id]['settings']['approved_senders'] = []
        self.get = _convert(self._cache[account_id]['settings'])
        self.account_id = account_id
        self.uid = self._cache[account_id]['uid']
        self.db_session = db_session
        self.default_locale = config.BABEL_DEFAULT_LOCALE

    def __new__(cls,  account_id, db_session=None, uid=None):
        if cls._singleton is None:
            cls._singleton = super(AccountSettings, cls).__new__(cls)
            config = ServiceConfig().config
            cls._cache = TTLCache(maxsize=config.CACHE_SIZE,
                                  ttl=config.REDIS_TTL)
        return cls._singleton

    def uncache(self):
        """Clear the cached settings for account_id"""
        self._cache.pop(self.account_id, None)

    @property
    def locale(self):
        """Returns the language for the uid if specified in the
        user's profile
        """
        try:
            return self.db_session.query(self.models.Profile).filter(
                self.models.Profile.uid == self.uid,
                self.models.Profile.item == 'lang').one().value
        except (NoResultFound, AttributeError):
            return None



"""const.py

Constants for API library
  Values that can be modified by a container restart belong in config.py

created 31-mar-2019 by richb@instantlinux.net
"""
from flask_babel import _


class Constants:
    """Constants for apicrud methods"""
    AUTH_INVITEE = 'invitee'
    CACHE_SIZE = 16384
    DEFAULT_AWS_REGION = 'us-east-2'
    DEFAULT_COUNTRY = u'US'
    DEFAULT_LANG = u'en_US'
    DEFAULT_WINDOW_TITLE = u'Example apicrud Application'
    LIB_MOD_SPATIALITE = ['/usr/lib/x86_64-linux-gnu/mod_spatialite.so',
                          '/usr/lib/mod_spatialite.so.7']
    MFA_BACKUP_CODELEN = 8
    MIME_IMAGE_TYPES = ("gif", "heic", "jpeg", "png", "svg")
    MIME_VIDEO_TYPES = ("mp4", "mpeg")
    PER_PAGE_DEFAULT = 100
    REDIS_TTL = 3600
    REGEX_EMAIL = r'[a-z0-9._%+-]+@[a-z0-9.-]+\.[a-z]{2,}'
    REGEX_PHONE = r'[0-9()+ -]{5,20}'
    SERVICE_CONFIG_FILE = 'service_config.yaml'
    TOTP_COOKIE = '__Secure-totp'


class i18nValues:
    """These values exist only to provide msgids for generating messages.pot"""

    contact_types = dict(
        email=_(u'email'), messenger=_(u'messenger'), sms=_(u'sms'),
        voice=_(u'voice'), whatsapp=_(u'whatsapp'))



"""totp_func

created 5-apr-2021 by richb@instantlinux.net
"""
from flask import g
from flask_babel import _
import logging
from hashlib import md5
import pyotp
from sqlalchemy.orm.exc import NoResultFound
from urllib.parse import urlparse

from .. import AccessControl, AESEncryptBin, Metrics, ServiceConfig
from ..const import Constants

# Accept previous value up to 30 seconds after OTP changes
TICKS = 1


def login(username, otp, redis_conn=None):
    """Log in with TOTP; a prior login must have succeeded and
    set up a session with the account_id and pendingotp auth, or
    with an apikey. If otp matches, this passes a sqlalchemy account
    record back for further processing by SessionAuth.login_accepted.

    Args:
      username (str): username or identity
      otp (str): 6 or 8-digit one-time password
      redis_conn (obj): connection to redis

    Returns:
      tuple: 4 items
        dict with error message
        http status (200 if OK)
        headers (Set-Cookie)
        account (sqlalchemy query object)
    """
    config = ServiceConfig().config
    acc = AccessControl()
    logmsg = dict(action='login_otp', username=username)
    query = g.db.query(ServiceConfig().models.Account)
    if acc.apikey_id:
        query = query.filter_by(name=username)
    else:
        query = query.filter_by(id=acc.account_id)
    try:
        account = query.one()
    except NoResultFound:
        logging.info(dict(message=u'account not found', **logmsg))
        return dict(message=_(u'access denied')), 403, None, None
    logmsg['identity'] = account.owner.identity
    if len(otp) == Constants.MFA_BACKUP_CODELEN:
        index, valid = 0, False
        codes = AESEncryptBin(config.DB_AES_SECRET).decrypt(
            account.backup_codes)
        otp_hash = md5(otp.encode()).digest()
        while index < len(codes):
            if otp_hash == codes[index:index + 16]:
                # snip the matching code
                account.backup_codes = codes[:index] + codes[index + 16:]
                g.db.commit()
                valid = True
                break
            # skip to the next 128-bit hash
            index += 16
        if not valid:
            logging.info(dict(backup_code='invalid', **logmsg))
            return dict(message=_(u'access denied')), 403, None, None
    elif not account.totp:
        msg = _(u'access denied')
        logging.warning(dict(message=msg, error='missing totp', **logmsg))
        return dict(message=msg), 403, None, None
    elif not pyotp.TOTP(account.totp_secret).verify(otp, valid_window=TICKS):
        Metrics().store('logins_fail_total')
        logging.info(dict(otp='invalid', **logmsg))
        g.db.commit()
        return dict(message=_(u'access denied')), 403, None, None
    # Block replay of same otp for next minute by setting a redis key
    result, error, status = None, 'otp replay', 403
    try:
        result = redis_conn.set(
            'totp:%s:%s' % (acc.account_id, otp), '', ex=60, nx=True)
    except Exception as ex:
        error, status = str(ex), 500
    if not result:
        msg = _(u'access denied')
        logging.warning(dict(message=msg, error=error, **logmsg))
        return dict(message=msg), status, None, None

    headers = None
    if config.LOGIN_MFA_COOKIE_LIMIT > 0:
        ses = g.session.create(
            acc.uid, [], acc=account.id, identity=account.owner.identity,
            ttl=config.LOGIN_MFA_COOKIE_LIMIT)
        hostname = urlparse(config.PUBLIC_URL).hostname
        path = '/' if acc.apikey_id else '%s/auth' % config.BASE_URL
        headers = {
            'Set-Cookie': '%s=%s; Path=%s; Max-Age=%s' % (
                config.LOGIN_MFA_COOKIE_NAME, ses.get('jti'), path,
                config.LOGIN_MFA_COOKIE_LIMIT)}
        if hostname != 'localhost':
            headers['Set-Cookie'] += (
                '; Domain=%s; HttpOnly; SameSite=Strict; Secure' % hostname)
    return {}, 200, headers, account



"""account controller

created 31-mar-2019 by richb@instantlinux.net
"""

from flask import g
from flask_babel import _
from sqlalchemy.orm.exc import NoResultFound

from apicrud import BasicCRUD, SessionAuth, state
from apicrud.auth import LocalUser


class AccountController(BasicCRUD):
    def __init__(self):
        super().__init__(resource='account')

    @staticmethod
    def create(body):
        return SessionAuth().account_add(body.get('identity'),
                                         body.get('uid'))

    @staticmethod
    def change_password(uid, body):
        """Change password

        Args:
          body (dict):
            Fields new_password, verify_password are required;
            either the old_password or a reset_token can be used
            to authorize the request.
        Returns:
          tuple: dict with account_id/uid/username, http response
        """
        return LocalUser().change_password(
            uid, body.get('new_password'), body.get('reset_token'),
            old_password=body.get('old_password'),
            verify_password=body.get('verify_password'))

    @staticmethod
    def get_password(uid):
        """Dummy get_password function is needed for react-admin's
        edit workflow (a get must precede put)

        Args:
          uid (str): User ID
        """
        self = state.controllers.get('account')
        try:
            account = g.db.query(self.model).filter_by(
                uid=uid, status='active').one()
        except NoResultFound:
            return dict(id=uid, message=_(u'account not found')), 404
        return dict(id=uid, username=account.name), 200

    @staticmethod
    def register(body):
        """Register a new account

        Args:
          body (dict):
            If forgot_password is set, send an email with reset token.
            Otherwise examine username, identity (email address) and
            name fields. Reject the new account if a duplicate identity
            or username already exists. Otherwise add the new account
            with Person and primary Contact objects. Finally send
            a confirmation email to the primary contact address.
        Returns:
          tuple: id of account created, and http status
        """
        if body.get('forgot_password'):
            return LocalUser().forgot_password(
                    body.get('identity'), body.get('username'))
        retval = LocalUser().register(
            body.get('identity'), body.get('username'), body.get('name'))
        if retval[1] > 201:
            return retval
        return SessionAuth().account_add(body['username'], retval[0]['uid'])



"""apikey

created 26-dec-2020 by richb@instantlinux.net
monolith broken out 18-apr-2021
"""
from datetime import timedelta
from flask import abort, g, request
from flask_babel import _
import logging
from sqlalchemy.orm.exc import NoResultFound

from .. import AccessControl, SessionAuth, Metrics
from ..utils import utcnow

APIKEYS = {}


class APIKey(SessionAuth):
    def access(self, apikey, otp=None):
        """Access using API key

        Args:
          apikey (str): the API key
          otp (str): 6 or 8-digit one-time password
        Returns:
          dict: uid, scopes (None if not authorized)
        """
        global APIKEYS
        acc = AccessControl()
        key_id, secret = apikey.split('.')
        if key_id not in APIKEYS or utcnow() > APIKEYS[key_id]['expires']:
            # Note - local caching in APIKEYS global var reduces database
            # queries on Accounts table
            uid, scopes = acc.apikey_verify(key_id, secret)
            if not uid:
                return None
            try:
                account = g.db.query(self.models.Account).filter_by(
                    uid=uid, status='active').one()
            except NoResultFound:
                logging.info(dict(action='api_key', uid=uid,
                                  message='account not active'))
                return None
            except Exception as ex:
                logging.error(dict(action='api_key', message=str(ex)))
                return None
            APIKEYS[key_id] = dict(
                account_id=account.id,
                expires=utcnow() + timedelta(seconds=self.config.REDIS_TTL),
                identity=account.owner.identity,
                roles=['admin', 'user'] if account.is_admin else ['user'],
                scopes=[item.name for item in scopes],
                totp=account.totp, uid=uid)
            APIKEYS[key_id]['roles'] += self.get_roles(uid)
        item = APIKEYS[key_id]
        uid = item['uid']
        if self.config.LOGIN_MFA_APIKEY:
            if self.config.LOGIN_MFA_REQUIRED and not item['totp']:
                abort(403, _(u'please configure MFA'))
            if item['totp'] and not self.totp_bypass(uid) and (
                    request.url_rule.rule.split('/')[3] != 'auth'):
                abort(401, _(u'no valid TOTP cookie, please re-auth'))
        duration = (self.login_admin_limit if 'admin' in item['roles']
                    else self.login_session_limit)
        token = acc.apikey_hash(secret)[:8]
        ses = None
        try:
            # Look for existing session in redis cache
            ses = g.session.get(
                uid, token, arg='auth', key_id=key_id).split(':')
        except AttributeError:
            ses = g.session.create(uid, item['roles'], acc=item['account_id'],
                                   identity=item['identity'],
                                   ttl=duration, key_id=key_id, nonce=token)
        if ses:
            return dict(uid=uid, scopes=item['scopes'])


def auth(apikey, required_scopes=None):
    """ API key authentication for openapi securitySchemes

    Args:
      apikey (str): the key
      required_scopes (list): permissions requested

    Returns:
      dict: uid if successful (None otherwise)
    """
    if len(apikey) > 96 or '.' not in apikey:
        return None
    retval = APIKey().access(apikey)
    if not retval:
        return retval
    if required_scopes and (set(required_scopes) & set(retval['scopes']) !=
                            set(required_scopes)):
        logging.info(dict(action='api_key', prefix=apikey.split('.')[0],
                          scopes=retval['scopes'], message='denied'))
        return None
    Metrics().store('api_key_auth_total')
    return retval



"""test.base

Mixin of shared functions for the TestBase class

created 10-oct-2019 by richb@instantlinux.net
"""
import base64
from collections import namedtuple
from contextlib import contextmanager
from dataclasses import dataclass
from flask import g
import json
import jwt
import os
import tempfile
import unittest
import yaml

from .. import database, service_config, ServiceConfig, SessionManager, \
    state, utils

test_globals = {}


class TestBaseMixin(object):

    def baseSetup(self):
        """Preliminary initialization: test code should call this once
        at initial pytest startup, after setting up a sqlite or other
        database and invoking initialize.app(). This function starts the
        flask test_client and makes global fixture vars available for
        the setUp classes.
        """
        unittest.mock.patch(
            'apicrud.service_registry.ServiceRegistry.update').start()
        test_globals['config'] = config = ServiceConfig().config
        test_globals['flask'] = test_globals['app'].test_client()
        with open(config.DB_SEED_FILE, 'rt', encoding='utf8') as f:
            records = yaml.safe_load(f)
        test_globals['constants'] = records.get('_constants')

    def classSetup(self):
        """Setup for each test. Instance vars are configured for the
        flask and redis emulators, a mock_messaging object for the
        messaging celery worker, and all the constants found in the
        data/db_fixtures.yaml file.
        """
        self.app = test_globals['app']
        self.config = test_globals['config']
        self.flask = test_globals['flask']
        self.redis = test_globals['redis']
        self.maxDiff = None
        self.base_url = self.config.BASE_URL
        self.credentials = {}
        self.authuser = None
        self.mock_messaging = state.func_send = unittest.mock.Mock()
        for item, val in test_globals['constants'].items():
            setattr(self, item, val)

    def authorize(self, username=None, password=None, apikey=None,
                  new_session=False, otp=None):
        """Authorize the specified user

        Args:
          username (str): login name
          password (str): password
          apikey (str): an API key
          new_session (bool): re-use existing session unless this is True
          otp (str): one-time password
        """
        if not username and not apikey:
            username = self.username
            password = self.password
        if apikey:
            response = self.call_endpoint(
                '/apikey?prefix=%s' % apikey[:8], 'get',
                extraheaders={self.config.HEADER_AUTH_APIKEY: apikey})
            if response.status_code == 200:
                self.apikey = apikey
            return response.status_code
        if username not in self.credentials or new_session or otp:
            response = self.call_endpoint('/auth', 'post', data=dict(
                username=username, password=password, otp=otp))
            if response.status_code != 201:
                return response.status_code
            tok = jwt.decode(response.get_json()['jwt_token'],
                             self.config.JWT_SECRET, algorithms=['HS256'])
            self.credentials[username] = dict(
                auth='Basic ' + base64.b64encode(
                    bytes(tok['sub'] + ':' + tok['jti'], 'utf-8')
                ).decode('utf-8'))
        self.authuser = username
        return 201

    def call_endpoint(self, endpoint, method, data=None, extraheaders={}):
        """call an endpoint with flask.g context

        params:
          endpoint(str): endpoint path (after base_url)
          method(str): get, post, put etc
          extraheaders(dict): additional headers
        returns:
          http response
        """

        base_url = '/config/v1' if endpoint == '/config' else self.base_url
        with self.app.test_request_context():
            g.db = database.get_session()
            g.session = SessionManager(redis_conn=self.redis)
            headers = extraheaders.copy()
            if hasattr(self, 'apikey'):
                headers[self.config.HEADER_AUTH_APIKEY] = self.apikey
                headers['Accept'] = 'application/json'
                headers['Content-Type'] = 'application/json'
            elif self.authuser:
                headers['Authorization'] = self.credentials[
                    self.authuser]['auth']
                headers['Accept'] = 'application/json'
                headers['Content-Type'] = 'application/json'
            if data:
                response = getattr(self.flask, method)(
                    base_url + endpoint, data=json.dumps(data),
                    headers=headers,
                    content_type='application/json')
            else:
                response = getattr(self.flask, method)(
                    base_url + endpoint, headers=headers)
            g.db.remove()
            return response

    @contextmanager
    def config_overrides(self, **kwargs):
        """Method to override ServiceConfig values, restoring upon
        completion. In normal production, the configuration is
        immutable until restart. This provides a test mechanism to override
        this security-design constraint. Example usage:

        with self.config_overrides(
                login_session_limit=10, login_attempts_max=1):
            ...tests...

        Args:
          kwargs: key/values to set
        Returns:
          (obj): the config object
        """
        config = ServiceConfig().config
        models = ServiceConfig().models
        saved_state = service_config.state.copy()
        saved_config = service_config.config._asdict()
        configfile = tempfile.mkstemp(prefix='_cfg')[1]

        def _restore():
            service_config.state = saved_state
            service_config.config = namedtuple('Struct', [
                key for key in saved_config.keys()])(*saved_config.values())
            os.remove(configfile)

        with open(configfile, 'w') as f:
            yaml.dump(kwargs, f)
        try:
            ret = ServiceConfig(
                file=configfile, reset=True,
                babel_translation_directories=(
                    config.BABEL_TRANSLATION_DIRECTORIES),
                db_schemas=config.DB_SCHEMAS,
                db_seed_file=config.DB_SEED_FILE,
                db_migrations=config.DB_MIGRATIONS,
                models=models, rbac_file=config.RBAC_FILE).config
        except AttributeError:
            _restore()
            raise

        try:
            yield ret
        finally:
            _restore()

    @contextmanager
    def scratch_account(self, username, name):
        """Set up a scratch account and login. This relies on
        a mock_messaging objected created in classSetup.

        Returns: obj
          id (str): account ID
          password (str): random password
          uid (str): ID of user
        """
        @dataclass
        class Return:
            id: str
            password: str
            uid: str

        account = dict(
            name=name, identity='%s@example.com' % username, username=username)
        pw = '@Xx%s' % utils.gen_id()
        password = dict(new_password=pw, verify_password=pw)

        response = self.call_endpoint('/account', 'post', data=account)
        self.assertEqual(response.status_code, 201)
        id = response.get_json()['id']
        uid = response.get_json()['uid']

        for call in self.mock_messaging.call_args_list:
            password['reset_token'] = call.kwargs.get('token')
        response = self.call_endpoint(
            '/account_password/%s' % uid, 'put', data=password)
        self.assertEqual(response.status_code, 200)
        status = self.authorize(username=account['username'], password=pw)
        self.assertEqual(status, 201)

        try:
            yield Return(id=id, password=pw, uid=uid)
        finally:
            self.authorize(username=self.admin_name, password=self.admin_pw)
            self.call_endpoint('/account/%s?force=true' % id, 'delete')
            self.assertEqual(response.status_code, 200)



"""send.py

created 18-apr-2019 by richb@instantlinux.net
"""

import celery
from flask_babel import _
import logging
import smtplib
from sqlalchemy.orm.exc import NoResultFound
import ssl

from ..account_settings import AccountSettings
from ..database import get_session
from ..exceptions import APIcrudSendError
from ..metrics import Metrics
from ..service_config import ServiceConfig
from ..utils import utcnow
from .format import MessageFormat


class Messaging(object):
    """
    External messaging

    Args:
        account_id (str): ID of logged-in account
        db_session (obj): open session to database
        settings (obj): AccountSettings for account
        smtp (obj): open session to SMTP smarthost
        ssl_context (obj): an SSL context (TLSv1_2)
    """
    ssl_context = ssl.SSLContext(ssl.PROTOCOL_TLSv1_2)

    def __init__(self, db_session=None, account_id=None, smtp=None,
                 settings=None):
        self.config = config = ServiceConfig().config
        self.db_session = db_session or get_session(
            scopefunc=celery.utils.threads.get_ident, db_url=config.DB_URL)
        self.models = ServiceConfig().models
        self.settings = settings or self._get_settings(
            self.models.Account, account_id=account_id)
        self.siteurl = self.settings.get.url
        self.smtp = smtp

    def __del__(self):
        if self.smtp:
            self.smtp.quit()
        self.db_session.remove()

    def send(self, frm=None, to=None, to_uid=None, template=None,
             attachments=[], **kwargs):
        """
        Send a message to one contact

        Args:
            frm (uid): person
            to (Contact): recipient (specific contact address)
            to_uid (uid): recipient (generic, use primary contact)
            template (str): jinja2 template name
            attachments (list): additional MIMEBase / MimeText objects
            kwargs: kv pairs
        Raises:
            APIcrudSendError
        """
        try:
            if to:
                to_contact = self.db_session.query(
                    self.models.Contact).filter_by(id=to).one()
            else:
                to_contact = self.db_session.query(
                    self.models.Contact).join(self.models.Person).filter(
                        self.models.Contact.info ==
                        self.models.Person.identity,
                        self.models.Contact.type == 'email',
                        self.models.Contact.status == 'active',
                        self.models.Person.id == to_uid).one().id
        except NoResultFound:
            msg = 'Recipient cid=%s, uid=%s not found' % (to, to_uid)
            logging.error(msg)
            raise APIcrudSendError(msg)
        logmsg = dict(action='send_contact', to_id=to, from_id=frm,
                      info=to_contact.info)
        if to_contact.status == 'disabled':
            logging.info(dict(status='disabled', **logmsg))
            return
        frm, from_contact = self._get_frm(frm, to_contact)
        if from_contact.info in self.settings.get.approved_senders:
            sender_email = from_contact.info
        else:
            sender_email = self.settings.get.sender_email
        if to_contact.type in ('email', 'sms'):
            if not self.smtp:
                self.smtp = self.smtp_session()
            logging.info('action=send_contact template=%s type=%s address=%s' %
                         (template, to_contact.type, to_contact.info))
            metrics = Metrics(uid=frm, db_session=self.db_session)
            if to_contact.type == 'sms':
                if (not metrics.store('sms_daily_total') or
                        not metrics.store('sms_monthly_total')):
                    msg = _(u'daily limit exceeded')
                    logging.info(dict(message=msg, **logmsg))
                    raise APIcrudSendError(msg)
                dest_email = (to_contact.info + '@' +
                              self.config.CARRIER_GATEWAYS[to_contact.carrier])
                body = MessageFormat(db_session=self.db_session,
                                     settings=self.settings).sms(
                    template, from_contact, to_contact, **kwargs)
            else:
                if (not metrics.store('email_daily_total') or
                        not metrics.store('email_monthly_total')):
                    msg = _(u'daily limit exceeded')
                    logging.info(dict(message=msg, **logmsg))
                    raise APIcrudSendError(msg)
                dest_email = to_contact.info
                body = MessageFormat(db_session=self.db_session,
                                     settings=self.settings).email(
                    template, from_contact, sender_email,
                    to_contact, attachments=attachments,
                    contact_id=to_contact.id, **kwargs).as_string()
            self.smtp.sendmail(sender_email, dest_email, body)
            to_contact.last_attempted = utcnow()
            self.db_session.add(to_contact)
        else:
            logging.warning(dict(message='Unsupported type',
                                 contact_type=to_contact.type, **logmsg))
            # self.db_session.remove()
            raise APIcrudSendError('unsupported type=%s' % to_contact.type)
        self.db_session.commit()

    def smtp_session(self):
        """Open an SMTP connection to the account's defined smtp_smarthost

        Args:
            settings (obj): settings object
        Raises:
            APIcrudSendError
        """
        session = smtplib.SMTP(self.settings.get.smtp_smarthost,
                               self.settings.get.smtp_port)
        session.starttls(context=self.ssl_context)
        if (hasattr(self.settings.get, 'smtp_credential_id') and
                self.settings.get.smtp_credential_id):
            try:
                cred = self.db_session.query(self.models.Credential).filter_by(
                    id=self.settings.get.smtp_credential_id).one()
                session.login(cred.key, cred.secret)
            except Exception as ex:
                logging.error(dict(action='send_contact', message=str(ex)))
                # self.db_session.remove()
                raise APIcrudSendError('Credential problem: %s' % str(ex))
        return session

    def _get_settings(self, model, account_id=None):
        if account_id is None:
            try:
                account_id = self.db_session.query(model).filter_by(
                    name='admin').one().id
            except NoResultFound:
                logging.error('action=_get_settings message="Missing admin"')
                raise APIcrudSendError('Missing admin')
        return AccountSettings(account_id, db_session=self.db_session)

    def _get_frm(self, frm, to_contact):
        """Get the from_contact record which matches the to_contact's
        type (sms, email etc)

        Args:
            frm (uid): the sender
            to_contact (obj): Contact record

        Returns: tuple
            frm (uid): sender uid (or administrator if noen)
            from_contact (obj): Contact record
        """
        logmsg = dict(action='_get_frm', from_id=frm, to=to_contact.info)
        if frm is None:
            frm = self.settings.get.administrator_id
        from_contact = None
        Contact = self.models.Contact
        try:
            from_contact = self.db_session.query(Contact).filter_by(
                uid=frm, type=to_contact.type).filter(
                    (Contact.status.in_(['active', 'unconfirmed']))).order_by(
                        Contact.rank).first()
        except Exception:
            # We don't care at all why this fails; next step catches error
            pass
        if not from_contact:
            try:
                from_contact = self.db_session.query(
                    Contact).filter_by(uid=frm).filter(
                        (Contact.status.in_(['active', 'unconfirmed']))
                        ).order_by(Contact.rank).first()
            except NoResultFound:
                # self.db_session.remove()
                raise APIcrudSendError('Contact for uid=%s not found' % frm)
            if from_contact:
                logging.warning(dict(message='missing contact type', **logmsg))
            else:
                logging.error(dict(message='garbled_from', **logmsg))
                # self.db_session.remove()
                raise APIcrudSendError('uid=%s cannot get From address' % frm)
        return frm, from_contact



"""contact controller

created 31-mar-2019 by richb@instantlinux.net
"""

from apicrud import BasicCRUD
from apicrud.messaging.confirmation import Confirmation


class ContactController(BasicCRUD):
    def __init__(self):
        super().__init__(resource='contact')

    @staticmethod
    def create(body):
        """after creating record, send confirmation message

        Args:
          body (dict): as defined in openapi.yaml schema
        """
        retval = super(ContactController, ContactController).create(body)
        if retval[1] == 201:
            result = Confirmation().request(retval[0]['id'])
            retval[0].update(result[0])
        return retval

    @staticmethod
    def update(id, body):
        """contact update - uses special-case function

        Args:
          id (str): resource ID
          body (dict): as defined in openapi.yaml
        """
        retval = super(ContactController, ContactController).update_contact(
            id, body)
        if retval[1] == 200:
            result = Confirmation().request(id)
            retval[0].update(result[0])
        return retval

    @staticmethod
    def confirmation_get(id):
        """Send a confirmation message via email (or as specified)

        Args:
          id (str): id of contact model
        """
        return Confirmation().request(id)

    @staticmethod
    def confirm(token):
        """Record a contact as fully confirmed

        Args:
          token (str): the token that was sent in confirmation message
        """
        # TODO parameterize this so new account gets pending role and
        # existing account gets person
        return Confirmation().confirm(token)



"""oauth2

created 26-mar-2021 by richb@instantlinux.net
"""
from flask import g
from flask_babel import _
import logging
from sqlalchemy.orm.exc import NoResultFound
import string

from .. import state
from ..database import db_abort
from ..metrics import Metrics
from ..session_auth import SessionAuth
from ..utils import gen_id, identity_normalize


class OAuth2(SessionAuth):
    """OAuth2 for Session Authorization
    """
    def __init__(self):
        super().__init__()
        self.func_send = state.func_send

    def callback(self, method, code=None, state=None):
        """Callback from 3rd-party OAuth2 provider auth

        Parse the response, look up the account based on email
        address, and pass control to SessionAuth.login_accepted

        Args:
          method (str): provider name, such as google
          code (str): validation code from provider
          state (str): provider state
        """
        client = self.oauth.create_client(method)
        logmsg = dict(action='oauth_callback', method=method)
        if not client:
            msg = _(u'login client missing')
            logging.error(dict(message=msg, **logmsg))
            return dict(message=msg), 500
        try:
            token = client.authorize_access_token(
                redirect_uri='%s%s/%s/%s' % (self.config.PUBLIC_URL,
                                             self.config.BASE_URL,
                                             'auth_callback', method))
        except Exception as ex:
            msg = _(u'openid client failure')
            logging.warning(dict(message=msg, error=str(ex), **logmsg))
            return dict(message=msg), 500
        if 'id_token' in token:
            user = client.parse_id_token(token)
        else:
            user = client.userinfo()
        if method == 'slack':
            user = user['user']
        identity = identity_normalize(user.get('email'))
        if not identity or '@' not in identity:
            logging.warning(dict(message='identity missing', usermeta=user,
                                 **logmsg))
            return dict(message=_(u'access denied')), 403
        try:
            contact = g.db.query(self.models.Contact).filter_by(
                    info=identity).one()
        except NoResultFound:
            return self._handle_unknown_user(method, user)
        except Exception as ex:
            return db_abort(str(ex), **logmsg)
        try:
            account = g.db.query(self.models.Account).join(
                self.models.Person).filter(
                    self.models.Person.identity == identity,
                    self.models.Account.status == 'active').one()
        except NoResultFound:
            account = None
        except Exception as ex:
            return db_abort(str(ex), **logmsg)
        if not account:
            try:
                account = g.db.query(self.models.Account).join(
                    self.models.Person).join(self.models.Contact).filter(
                        self.models.Contact.info == identity,
                        self.models.Contact.type == 'email').one()
            except NoResultFound:
                logging.warning(dict(message='lookup failed', **logmsg))
                return dict(message=_(u'access denied')), 403
            except Exception as ex:
                return db_abort(str(ex), **logmsg)
        if account.owner.status != 'active' or contact.status == 'disabled':
            logging.info(dict(message='disabled account', **logmsg))
            return dict(message=_(u'access denied')), 403
        return self.login_accepted(account.name, account, method)

    def _handle_unknown_user(self, method, usermeta):
        """Handle unknown external user access based on configured
        LOGIN_EXTERNAL_POLICY.

        Args:
          method (str): login method, as in 'google'
          usermeta (dict): the metadata object from external provider
        """
        identity = identity_normalize(usermeta.get('email'))
        name = usermeta.get('name')
        picture = usermeta.get('picture') or usermeta.get('avatar_url')
        if not picture and usermeta.get('gravatar_id'):
            picture = 'https://2.gravatar.com/avatar/%s' % usermeta.get(
                'gravatar_id')
        username = '%s-%s' % (
            identity.split('@')[0][:15],
            gen_id(length=6, prefix='',
                   chars=string.digits + string.ascii_lowercase))
        logmsg = dict(action='register', method=method, identity=identity)

        if self.config.LOGIN_EXTERNAL_POLICY == 'closed':
            msg = _(u'not valid')
            logging.info(dict(msg=msg, **logmsg))
            Metrics().store('logins_fail_total')
            return dict(message=msg), 403
        elif self.config.LOGIN_EXTERNAL_POLICY == 'auto':
            ret = self.register(identity, username, name, picture=picture,
                                template=None)
            if ret[1] != 200:
                return ret
            ret = self.account_add(username, uid=ret[0]['uid'])
            if ret[1] != 201:
                return ret
            account = g.db.query(self.models.Account).filter_by(
                id=ret[0]['id']).one()
            logging.info(dict(usermeta=usermeta, **logmsg))
            return self.login_accepted(username, account, method)
        elif self.config.LOGIN_EXTERNAL_POLICY == 'open':
            ret = self.register(identity, username, name, picture=picture)
            if ret[1] != 200:
                return ret
            ret2 = self.account_add(username, uid=ret[0]['uid'])
            logging.info(dict(usermeta=usermeta, **logmsg))
            return ret2 if ret2 == 201 else ret

        # TODO LOGIN_EXTERNAL_POLICY == 'onrequest'
        return dict(message='unimplemented'), 403



"""Database model base

created 26-mar-2019 by richb@instantlinux.net

license: lgpl-2.1
"""

from sqlalchemy.ext.declarative import declarative_base

Base = declarative_base()
aes_secret = None


def get_aes_secret():
    return aes_secret


class AsDictMixin(object):

    def as_dict(self):
        """Returns a serializable dict from an instance of the model

        Each table column is returned in the dict, except for any
        listed in a __rest_exclude__ tuple

        A list of related records' ids can be attached by defining
        one or more resource names in a __rest_related__ tuple
        """

        retval = self.__dict__.copy()
        if hasattr(self, '__rest_exclude__'):
            for col in self.__rest_exclude__:
                del(retval[col])
        if hasattr(self, '__rest_related__'):
            for key in self.__rest_related__:
                retval[key] = [rec.id for rec in getattr(self, key)]
        if hasattr(self, '__rest_hybrid__'):
            for key in self.__rest_hybrid__:
                retval[key] = getattr(self, key)
        retval.pop('_sa_instance_state', None)
        return retval



"""local_user

created 26-mar-2020 by richb@instantlinux.net
monolith broken out 6-apr-2021
"""
from flask import g
from flask_babel import _
import logging
from passlib.hash import sha256_crypt
import redis.exceptions
from sqlalchemy import or_
from sqlalchemy.orm.exc import NoResultFound
import string

from .. import SessionAuth
from ..utils import identity_normalize, utcnow
from ..messaging.confirmation import Confirmation


class LocalUser(SessionAuth):
    """Manage local user accounts
    """
    def __init__(self):
        super().__init__()

    def change_password(self, uid, new_password, reset_token,
                        old_password=None, verify_password=None):
        """Update a user's password, applying complexity rules; must
        specify either the old password or a reset token

        Args:
          uid (str): User ID
          new_password (str): the new passphrase
          reset_token (str): a token retrieved from Confirmation.request
          old_password (str): the old passphrase

        Returns:
          tuple: dict with account_id/uid/username, http response
        """

        if not new_password or new_password != verify_password:
            return dict(message=_(u'passwords do not match')), 405
        try:
            account = g.db.query(self.models.Account).filter_by(
                uid=uid, status='active').one()
        except NoResultFound:
            return dict(uid=uid, message=_(u'account not found')), 404
        logmsg = dict(action='change_password', resource='account',
                      username=account.name, uid=uid)

        if self._password_weak(new_password):
            return dict(message=_(u'rejected weak password')), 405
        if old_password:
            if not sha256_crypt.verify(old_password, account.password):
                msg = 'invalid credential'
                logging.warning(dict(message=msg, **logmsg))
                return dict(username=account.name, message=msg), 403
        else:
            retval = Confirmation().confirm(reset_token,
                                            clear_session=True)
            if retval[1] != 200:
                msg = _(u'invalid token')
                logging.warning(dict(message=msg, **logmsg))
                return dict(username=account.name, message=msg), 405
        account.password = sha256_crypt.hash(new_password)
        account.password_must_change = False
        g.db.commit()
        logging.info(dict(message=_(u'changed'), **logmsg))
        return dict(id=account.id, uid=uid, username=account.name), 200

    def forgot_password(self, identity, username, template='password_reset'):
        """Trigger Confirmation.request; specify either the username
        or email address. For security, administrators are not allowed
        to use this feature.

        Args:
          identity (str): account's primary identity, usually an email
          username (str): account's username
          template (str): template for message (confirming new user)
        Returns:
          tuple: the Confirmation.request dict and http response
        """
        identity = identity_normalize(identity)
        logmsg = dict(action='forgot_password', identity=identity,
                      username=username)
        try:
            account = g.db.query(self.models.Account).join(
                self.models.Person).filter(
                    self.models.Account.is_admin == 0,
                    self.models.Account.status == 'active',
                or_(
                    self.models.Account.name == username,
                    self.models.Person.identity == identity)).one()
            id = g.db.query(self.models.Contact).filter_by(
                type='email', info=account.owner.identity).one().id
        except NoResultFound:
            logging.warning(dict(message='not found', **logmsg))
            if identity and '@' in identity:
                # TODO send reset_invalid message instead of rejection
                return dict(message=_(u'username or email not found')), 404
            return dict(message=_(u'username or email not found')), 404
        logging.info(logmsg)
        return Confirmation().request(id, template=template)

    @staticmethod
    def _password_weak(password):
        """Validate password complexity

        Args:
          password (str): Password
        Returns:
          bool: True if at least 3 different types of characters
        """
        return (not set(password).intersection(string.ascii_lowercase),
                not set(password).intersection(string.ascii_uppercase),
                not set(password).intersection(string.digits),
                not set(password).intersection(string.punctuation)
                ).count(True) > 1


# Support openapi securitySchemes: basic_auth endpoint
def basic(username, password, required_scopes=None):
    """This is a modified basic-auth validation function. The auth login
    controller method generates a random 8-byte token, stores it in
    the session_manager object as token_auth, and sends it to javascript
    authProvider. The dataProvider must send it back to us as
    basic-auth (base64-encoded).

    Vulnerable to session-hijacking if auth packets aren't encrypted
    end to end, but "good enough" until OAuth2 effort is completed.

    Implemented because of https://github.com/zalando/connexion/issues/791

    Args:
      username (str): Session UID
      password (str): Session token
      required_scopes (list): not used
    Returns:
      dict: uid with the username passed in
    """
    session = g.session.get(username, password)
    token_auth = 'ok'
    if session is None or 'jti' not in session:
        token_auth = 'missing'
    elif session['jti'] != password or session['sub'] != username:
        token_auth = 'invalid'
    elif 'exp' in session and utcnow().strftime('%s') > session['exp']:
        token_auth = 'expired'

    if token_auth != 'ok':
        logging.info('action=logout username=%s token_auth=%s' % (
            username, token_auth))
        if session:
            try:
                g.session.delete(username, password)
            except redis.exceptions.ConnectionError as ex:
                logging.error(dict(action='logout', message=str(ex)))
        return None
    return dict(uid=username)



"""credential.py

Credential storage

created 13-mar-2019 by richb@instantlinux.net
"""

from apicrud import BasicCRUD


class CredentialController(BasicCRUD):
    def __init__(self):
        super().__init__(resource='credential')



"""exceptions.py

created 21-jan-2021 by richb@instantlinux.net
"""


class APIcrudException(Exception):
    pass


class APIcrudFormatError(APIcrudException):
    pass


class APIcrudSendError(APIcrudException):
    pass


class MediaUploadError(Exception):
    pass



"""basic_crud.py

created 31-mar-2019 by richb@instantlinux.net
"""

import base64
from connexion import NoContent
from datetime import datetime, timedelta
from flask import g, request
from flask_babel import _
import json
import logging
import re
from sqlalchemy import asc, desc
from sqlalchemy.exc import IntegrityError, InvalidRequestError
from sqlalchemy.orm.exc import NoResultFound

from .access import AccessControl
from .account_settings import AccountSettings
from .const import Constants
from .database import db_abort
from .grants import Grants
from .service_config import ServiceConfig
from . import geocode, state, utils


class BasicCRUD(object):
    """Controller base class

    Create/Read/Update/Delete/find controller operations.

    This class provides permission-based, paginated access to database
    models behind your application's endpoints. Most endpoints need no
    boilerplate code, and can inherit these functions directly. Some
    endpoints only need a few lines of code before or after inheriting
    these functions. You can always write your own custom function for
    special-case endpoints.

    Args:
      resource (str): a resource name (endpoint prefix)
      model (obj): the model corresponding to the resource
    """

    def __init__(self, resource=None, model=None):
        self.models = state.models
        self.resource = resource
        if self.resource not in state.controllers:
            if model:
                self.model = model
            else:
                self.model = getattr(self.models, resource.capitalize())
            state.controllers[self.resource] = self

    @staticmethod
    def create(body, id_prefix='x-', limit_related={}):
        """Controller for POST endpoints. This method assigns a new
        object ID, sets the _created_ timestamp, evaluates user's
        permissions, adds a default category_id if the model has
        this attribute, and inserts a row to the back-end database.

        Args:
          body (dict): resource fields as defined by openapi.yaml schema
          id_prefix (str): generated objects will be assigned a random 10-
            to 16-character ID; you can set a unique prefix if desired
          limit_related (dict): limits on number of related records,
            keyed by relationship name
        Returns:
          tuple:
            first element is a dict with the id, second element is
            response code (201 on success)
        """

        self = state.controllers[request.url_rule.rule.split('/')[3]]
        # TODO reject extraneous keys beyond those in the openapi path spec
        if self.resource == 'contact':
            retval = self._create_contact(body)
            if retval[1] != 201:
                return retval
        acc = AccessControl(model=self.model)
        logmsg = dict(action='create', account_id=acc.account_id,
                      resource=self.resource, ident=acc.identity)
        if 'id' in body:
            return dict(message='id is a read-only property',
                        title='Bad Request'), 405
        body['id'] = id = utils.gen_id(prefix=id_prefix)
        body['created'] = utils.utcnow()
        if body.get('expires'):
            try:
                body['expires'] = self._fromdate(body['expires'])
            except Exception:
                return dict(message=_(u'invalid date')), 405
        if hasattr(self.model, 'uid') and not body.get('uid'):
            body['uid'] = acc.uid
        uid = logmsg['uid'] = body.get('uid')
        if not acc.with_permission('c', new_uid=uid,
                                   membership=acc.primary_resource,
                                   id=body.get('event_id')):
            logging.warning(dict(message='access denied', **logmsg))
            return dict(message=_(u'access denied')), 403
        ret_info = {}
        logging.info(dict(id=id, name=body.get('name'), **logmsg))
        if not body.get('category_id') and hasattr(self.model, 'category_id'):
            if acc.account_id:
                body['category_id'] = AccountSettings(
                    acc.account_id, g.db).get.category_id
            elif body.get('event_id'):
                body['category_id'] = g.db.query(self.models.Event).filter_by(
                    id=body['event_id']).one().category_id
            else:
                logging.warning(dict(message='unexpected no creds', **logmsg))
                return dict(message=_(u'access denied')), 403
        if hasattr(self.model, 'status'):
            body['status'] = body.get('status', 'active')

        grant = (self.resource if self.resource.endswith('s')
                 else self.resource + 's')
        limit = Grants().get(grant, uid=uid)
        if grant in ServiceConfig().config.DEFAULT_GRANTS and g.db.query(
                self.model).filter_by(uid=uid).count() >= limit:
            msg = _('user limit exceeded')
            logging.info(dict(message=msg, allowed=limit, **logmsg))
            return dict(message=msg, allowed=limit), 405
        related_items = None
        if hasattr(self.model, '__rest_related__'):
            related_items = {
                item: body[item] for item in self.model.__rest_related__
                if item in body}
            for related in self.model.__rest_related__:
                if (related in body and related in limit_related and
                        len(body[related]) > limit_related[related]):
                    logging.warning(dict(message='limit exceeded',
                                         field=related, **logmsg))
                    return dict(message=_(u'limit exceeded')), 405
                body.pop(related, None)
        if self.resource == 'apikey':
            body['prefix'], secret, body['hashvalue'] = acc.apikey_create()
            ret_info = dict(apikey=body['prefix'] + '.' + secret,
                            name=body['name'])
        try:
            record = self.model(**body)
        except (AttributeError, TypeError) as ex:
            logging.warning(dict(message=str(ex), **logmsg))
            return dict(message=str(ex)), 405
        g.db.add(record)
        try:
            if related_items:
                g.db.commit()
                for related, records in related_items.items():
                    self._update_related(id, related, records)
            g.db.commit()
        except IntegrityError as ex:
            message = 'duplicate or other conflict'
            logging.warning(dict(message=message, error=str(ex), **logmsg))
            return dict(message=message, data=str(body)), 405
        except Exception as ex:
            return db_abort(str(ex), rollback=True, **logmsg)
        return dict(id=id, **ret_info), 201

    @staticmethod
    def get(id):
        """Controller for GET endpoints. This method evaluates
        privacy settings against the user's permissions, looks up
        category, owner and geocode values, and fetches the object
        from back-end database.

        Args:
          id (str): ID of the desired resource
        Returns:
          tuple:
            first element is a dict with the object or error
            message, second element is response code (200 on success)
        """

        self = state.controllers[request.url_rule.rule.split('/')[3]]
        acc = AccessControl(model=self.model)
        try:
            query = self.db_get(id)
            record = query.one()
        except NoResultFound as ex:
            msg = _(u'not found') if 'No row was found' in str(ex) else str(ex)
            return dict(id=id, message=msg), 404

        retval = record.as_dict()
        if hasattr(self.model, 'uid') and hasattr(self.model, 'owner'):
            retval['owner'] = record.owner.name
        if hasattr(self.model, 'category_id') and retval['category_id']:
            retval['category'] = record.category.name
        if 'modified' in retval and not retval['modified']:
            del(retval['modified'])
        eid = None if hasattr(self.model, 'event_id') else (
            acc.auth_ids[acc.primary_resource] or [None])[0]
        retval['rbac'] = ''.join(sorted(list(
            acc.rbac_permissions(query=query, membership=acc.primary_resource,
                                 id=eid) - set('c'))))
        if 'geolat' in retval:
            access = 'r' if 'r' in retval['rbac'] else None
            if retval['privacy'] == 'secret' and not access:
                return dict(message=_(u'access denied')), 403
            return geocode.with_privacy(retval, access), 200
        if 'r' in retval['rbac']:
            logging.info(dict(
                action='get', resource=self.resource, id=id,
                ident=acc.identity, duration=utils.req_duration()))
            return retval, 200
        else:
            return dict(message=_(u'access denied'), id=id), 403

    @staticmethod
    def update(id, body, access='u', limit_related={}):
        """Controller for PUT endpoints. This method looks for an existing
        record, evaluates user's permissions, and updates the row in
        the back-end database.

        Args:
          body (dict): fields to be updated
          access (str): access-level required for RBAC evaluation
          limit_related (dict): limits on number of related records,
            indexed by relationship name
        Returns:
          dict:
            first element is a dict with the id, second element is
            response code (200 on success)
        """
        if 'id' in body and body['id'] != id:
            return dict(message='id is a read-only property',
                        title='Bad Request'), 405
        self = state.controllers[request.url_rule.rule.split('/')[3]]
        body['modified'] = utils.utcnow()
        if body.get('expires'):
            try:
                body['expires'] = self._fromdate(body['expires'])
            except Exception:
                return dict(message=_(u'invalid date')), 405
        if 'totp' in body and not body.pop('totp'):
            body['totp_secret'] = None
        if hasattr(self.model, '__rest_related__'):
            for related in self.model.__rest_related__:
                if related in body:
                    ret = self._update_related(
                        id, related, body.pop(related, None),
                        limit=(limit_related[related]
                               if related in limit_related else None))
                    if ret[1] != 200:
                        return ret
        logmsg = dict(action='update', resource=self.resource, id=id,
                      ident=AccessControl().identity)
        try:
            query = g.db.query(self.model).filter_by(id=id)
            if not AccessControl(
                    model=self.model).with_permission(
                        access, query=query):
                return dict(message=_(u'access denied'), id=id), 403
            current = query.one().as_dict()
            query.update(body)
            g.db.commit()
        except IntegrityError as ex:
            message = _(u'duplicate or other conflict')
            logging.warning(dict(message=message, error=str(ex), **logmsg))
            return dict(message=message, data=str(ex)), 405
        except Exception as ex:
            return db_abort(str(ex), rollback=True, **logmsg)

        updated = {key: body[key] for key, val in body.items()
                   if key in current and current[key] != val and
                   key not in ('name', 'modified')}
        logging.info(dict(name=body.get('name'), duration=utils.req_duration(),
                          **logmsg, **updated))
        return dict(id=id, message=_(u'updated')), 200

    @staticmethod
    def delete(ids, force=False):
        """Controller for DELETE endpoints. This method looks for existing
        records, evaluates user's permissions, and updates or removes
        rows in the back-end database.

        Args:
          ids (list of str): record IDs to be flagged for removal
          force (bool): flag for removal if false; remove data if true
        Returns:
          tuple:
            first element is a dict with the id, second element is
            response code (200 on success)
        """

        # TODO - update auth if model could affect any session's auth
        self = state.controllers[request.url_rule.rule.split('/')[3]]
        logmsg = dict(action='delete', resource=self.resource,
                      account_id=AccessControl().account_id,
                      ident=AccessControl().identity)
        errors = 0
        count = 0
        for id in ids:
            try:
                query = g.db.query(self.model).filter_by(id=id)
                if not AccessControl(
                        model=self.model).with_permission(
                            'd', query=query):
                    return dict(message=_(u'access denied'), id=id), 403
                if force:
                    if query.delete():
                        logging.info(dict(id=id, **logmsg))
                    else:
                        logging.info(dict(id=id, msg='query failed', **logmsg))
                        errors += 1
                else:
                    logging.info(dict(id=id, status='disabled', **logmsg))
                    query.update(dict(status='disabled'))
            except Exception as ex:
                return db_abort(str(ex), **logmsg)
            count += 1
        try:
            g.db.commit()
        except Exception as ex:
            return db_abort(str(ex), rollback=True, **logmsg)
        logging.info(dict(count=count, ids=ids, **logmsg))
        return NoContent, 404 if errors else 204

    @staticmethod
    def find(**kwargs):
        """Find records which match query parameters passed from
        connexion by name, in a dictionary that also includes user
        and token info

        Args:
          cursor_next (str): pagination token to fetch subsequent records
          filter (dict): field/value pairs to query (simple queries
              only, with string or list matching; or * for any)
          limit (int): max records to fetch
          offset (int): old-style pagination starting offset
          sort (str): <field>[:{asc|desc}]
          status (str): value is added to filter
        Returns:
          dict: items (list), count(int), cursor_next (str)
        """

        self = state.controllers[request.url_rule.rule.split('/')[3]]
        acc = AccessControl(model=self.model)
        logmsg = dict(action='find', ident=acc.identity,
                      resource=self.resource)
        # TODO this conditional is a band-aid to give status 403 instead of
        # 200 with empty item list for anonymous / restricted users
        if self.resource != 'account' and not acc.with_permission(
               'r', new_uid=acc.uid):
            return dict(message=_(u'access denied')), 403
        conditions = {item: value for item, value in kwargs.items()
                      if item in ('status')}
        if 'cursor_next' in kwargs:
            offset = int(self._fromb64(kwargs['cursor_next']).split(':')[1])
        elif 'offset' in kwargs:
            offset = int(kwargs['offset'])
        else:
            offset = 0
        sort = kwargs.get('sort',
                          'name' if hasattr(self.model, 'name') else 'id')
        if ':' in sort:
            sort, dir = sort.split(':')
            sortdir = desc if dir == 'desc' else asc
        else:
            sortdir = asc
        query = g.db.query(self.model)
        limit = int(kwargs.get('limit', Constants.PER_PAGE_DEFAULT))
        try:
            filter = json.loads(kwargs.get('filter', '{}'))
        except json.decoder.JSONDecodeError as ex:
            msg = _(u'invalid filter specified') + '=%s' % str(ex)
            logging.error(dict(message=msg, **logmsg))
            return dict(message=msg), 405
        if 'id' in filter:
            # TODO: this seems a bit hackish, try to remember what
            # use-case this supports. The del() use-case is to
            # enable get:guest?filter{id:null,event_id=foo}
            if not filter.get('id'):
                if len(filter) == 1:
                    return dict(count=0, items=[]), 200
                else:
                    del(filter['id'])
        elif 'filter' in kwargs and not filter:
            return dict(count=0, items=[]), 200
        if filter.get('account_id') and not hasattr(self.model, 'account_id'):
            try:
                account = g.db.query(self.models.Account).filter_by(
                    id=filter['account_id']).one()
            except NoResultFound:
                return dict(count=0, items=[],
                            message=_(u'not found')), 404
            filter['uid'] = account.uid
            filter.pop('account_id')
        for key in filter.copy():
            if filter[key] == '*':
                filter.pop(key)
            elif (filter[key] and type(filter[key]) is str and '%'
                  in filter[key]):
                query = query.filter(getattr(self.model, key).like(
                    filter[key]))
                filter.pop(key)
            elif type(filter[key]) is list:
                items = filter.pop(key)
                if len(items) == 1:
                    filter[key] = items[0]
                else:
                    query = query.filter(getattr(self.model, key).in_(items))
        if hasattr(self.model, 'starts'):
            if filter.get('prev'):
                query = query.filter(self.model.starts < utils.utcnow())
            else:
                # Filter out records that start more than 12 hours ago
                query = query.filter(self.model.starts >
                                     utils.utcnow() - timedelta(hours=12))
            filter.pop('prev', None)
        try:
            query = query.filter_by(**filter).order_by(
                sortdir(getattr(self.model, sort)))
        except InvalidRequestError as ex:
            logging.warning(dict(message=str(ex), **logmsg))
            return dict(message=_(u'invalid filter specified')), 405
        if 'status' in conditions and conditions['status'] == 'disabled':
            query = query.filter(self.model.status == 'disabled')
        else:
            query = query.filter(self.model.status != 'disabled')
        query = acc.with_filter(query)
        try:
            results = query.slice(offset, offset + limit + 1).all()
            retval = dict(items=[], count=query.count())
        except Exception as ex:
            return db_abort(str(ex), **logmsg)
        count = 0
        for result in results[:limit]:
            record = result.as_dict()
            if hasattr(self.model, 'owner'):
                record['owner'] = result.owner.name
            # TODO - find a way to avoid this, causes KeyError on lazy-load
            if record.get('category_id'):
                record['category'] = result.category.name
            # TODO - takes 30ms per record for rbac: optimize or redesign
            record['rbac'] = ''.join(sorted(list(
                acc.rbac_permissions(owner_uid=record.get('uid'),
                                     membership=acc.primary_resource,
                                     id=record.get('id'),
                                     privacy=record.get('privacy'))
                - set('c'))))
            retval['items'].append(record)
            count += 1
        if len(results) > limit:
            retval['cursor_next'] = self._tob64('cursor:%d' % (offset + limit))
        elif count < retval['count'] and count < limit and offset == 0:
            # TODO find a way to get query.count() to return accurate value
            retval['count'] = count
        logging.info(dict(
            offset=offset, limit=limit,
            duration=utils.req_duration(), **conditions, **logmsg))
        return retval, 200

    def _create_contact(self, body):
        """Perform pre-checks against fields for contact resource
        prior to rest of contact-create

        Args:
            body (dict): as defined in openapi.yaml schema
        """
        logmsg = dict(action='create', resource='contact',
                      uid=body.get('uid'))
        acc = AccessControl()
        if body.get('type') == 'sms':
            if body.get('carrier') is None:
                return dict(message=_(u'carrier required for sms')), 405
            elif not re.match(Constants.REGEX_PHONE, body['info']):
                return dict(message=_(u'invalid mobile number')), 405
            body['info'] = re.sub('[- ()]', '', body['info'])
        elif body.get('type') == 'email':
            body['info'] = utils.identity_normalize(body['info'].strip())
            if not re.match(Constants.REGEX_EMAIL, body['info']):
                return dict(message=_(u'invalid email address')), 405
        elif 'type' in body and body.get('type') not in ['sms', 'email']:
            return dict(message='contact type not yet supported'), 405
        if not body.get('rank'):
            try:
                count = g.db.query(self.models.Contact).filter_by(
                    uid=body.get('uid'), type=body.get('type')).count()
            except Exception as ex:
                return db_abort(str(ex), **logmsg)
            body['rank'] = count + 1
        if body.get('uid', acc.uid) != acc.uid and 'admin' not in acc.auth:
            logging.warning(dict(message=_(u'access denied'), **logmsg))
            return dict(message=_(u'access denied')), 403
        if not body.get('status'):
            body['status'] = 'unconfirmed'
        return dict(status='ok'), 201

    @staticmethod
    def update_contact(id, body):
        """This is a special-case function for the contact-update resource

        - validate sms carrier
        - keep person identity in sync with primary contact

        Args:
          id (str): resource ID
          body (dict): as defined in openapi.yaml
        """

        logmsg = dict(action='update', id=id, info=body.get('info'))
        if body.get('type') == 'sms':
            if body.get('carrier') is None:
                return dict(message=_(u'carrier required for sms')), 405
            if not re.match(Constants.REGEX_PHONE, body['info']):
                return dict(message=_(u'invalid mobile number')), 405
            body['info'] = re.sub('[- ()]', '', body['info'])
        elif body.get('type') == 'email':
            body['info'] = utils.identity_normalize(body['info'].strip())
            if not re.match(Constants.REGEX_EMAIL, body['info']):
                return dict(message=_(u'invalid email address')), 405
        if 'id' in body and body['id'] != id:
            return dict(message='id is a read-only property',
                        title='Bad Request'), 405
        self = state.controllers[request.url_rule.rule.split('/')[3]]
        body['modified'] = utils.utcnow()
        try:
            query = g.db.query(self.model).filter_by(id=id)
            if not AccessControl(model=self.model).with_permission(
                    'u', query=query):
                return dict(message=_(u'access denied'), id=id), 403
            prev_identity = query.one().info
            if body.get('status') != 'disabled':
                body['status'] = 'unconfirmed'
            query.update(body)
            try:
                # If updating primary contact, also update identity
                primary = g.db.query(self.models.Person).filter_by(
                    identity=prev_identity).one()
                logging.info(dict(
                    resource='person', previous=prev_identity, **logmsg))
                primary.identity = body.get('info')
            except NoResultFound:
                pass
            g.db.commit()
            logging.info(dict(resource=self.resource, **logmsg))
        except Exception as ex:
            logging.warning(dict(message=str(ex), **logmsg))
            g.db.rollback()
            return dict(message=_(u'conflict with existing')), 405

        return dict(id=id, message=_(u'updated')), 200

    def _update_related(self, id, attr, related_ids, limit=None):
        """Update the list of records in the many-to-many association
        model for one column in the record. Adds and/or removes entries
        from the association table to make the list match related_ids.

        Args:
          id (str): ID of record in current model
          attr (str): relationship attribute name
          related_ids (list): list of existing records in related table
          limit (int): maximum number of allowed records

        Returns: tuple
          flask-compatible status dict, http return status
        """
        if limit and len(related_ids) > limit:
            msg = 'Limit exceeded'
            logging.warn(dict(
                action='update', resource=self.resource, allowed=limit,
                field=attr, uid=AccessControl().uid, message=msg))
            return dict(message=msg, allowed=limit), 405
        if attr == 'members':
            model = getattr(self.models, 'Person')
        else:
            model = getattr(self.models, attr.capitalize().strip('s'))
        query = g.db.query(self.model).filter_by(id=id).one()
        current = [item.id for item in getattr(query, attr)]
        for missing_member in set(related_ids) - set(current):
            getattr(query, attr).append(g.db.query(model).filter_by(
                id=missing_member).one())
        g.db.flush()
        for removed_member in set(current) - set(related_ids):
            getattr(query, attr).remove(g.db.query(model).filter_by(
                id=removed_member).one())
        g.db.commit()
        return dict(id=id, status='ok', items=len(related_ids)), 200

    @staticmethod
    def _tob64(text):
        return base64.b64encode(bytes(text, 'utf8')).decode('ascii')

    @staticmethod
    def _fromb64(text):
        return base64.b64decode(bytes(text, 'utf8')).decode('ascii')

    @staticmethod
    def _fromdate(date_string):
        """Convert a serialized ISO 8601 date-string to datetime format;
        handle variants where milliseconds are specified or hour/min/sec
        are omitted

        Args:
          date_string (str): ISO 8601 string like 2021-06-02T22:15:00Z
        Returns:
          datetime object
        """
        exp = date_string
        if re.match('^2[0-9]{3}-((0[1-9])|(1[0-2]))-([0-2][1-9]|3[0-1])$',
                    exp):
            exp += 'T23:59:59Z'
        try:
            expires = datetime.strptime(exp, '%Y-%m-%dT%H:%M:%S.%fZ')
        except ValueError:
            expires = datetime.strptime(exp, '%Y-%m-%dT%H:%M:%SZ')
        return expires

    def db_get(self, id):
        """Activate a SQLalchemy query object for the specified ID in
        the current model

        Args:
          id (str): object ID
        Returns:
          obj: query object
        """
        return g.db.query(self.model).filter_by(id=id)



"""confirmation.py

Confirmation
  Handle the request and confirm actions for a contact.

created 4-apr-2020 by richb@instantlinux.net
"""

from flask import g
from itsdangerous import URLSafeTimedSerializer
import logging
import random
from sqlalchemy.orm.exc import NoResultFound
import string

from .. import state
from ..service_config import ServiceConfig


class Confirmation:
    """Functions for confirming ownership of online email/contact info;
    class attributes are passed through the service config.

    Attributes:
      token_salt (str): salt value for serializer token
      token_secret (str): a secret to validate upon confirmation
      token_timeout (int): seconds until expiration
    """

    def __init__(self):
        config = ServiceConfig().config
        self.func_send = state.func_send
        self.models = ServiceConfig().models
        self.token_salt = config.TOKEN_SALT
        self.token_secret = config.TOKEN_SECRET
        self.token_timeout = config.TOKEN_TIMEOUT

    def request(self, id, ttl=None, template='contact_add'):
        """Generate a confirmation token valid for a given period,
        and send it to the contact identified by id

        Args:
          id (str): record ID in contacts table of database
          ttl (int): how many seconds before token expires
          template (str): jinja2 template name contact-add message

        Returns:
          tuple:
            first item is dict with token and contact-ID, second is
            the http response code
        """

        ttl = ttl or self.token_timeout
        contact = g.db.query(self.models.Contact).filter_by(id=id).one()
        if not contact:
            return dict(id=id, message='Not Found'), 404
        serializer = URLSafeTimedSerializer(self.token_secret)
        if contact.type == u'sms':
            # For SMS, need to use a short token (less secure)
            nonce = '1'
        else:
            nonce = ''.join([random.choice(string.ascii_letters)
                             for i in range(8)])
        token = serializer.dumps('%s:%s:%s' % (
            contact.id, contact.info, nonce), salt=self.token_salt)
        g.session.create(contact.id, [], param=token, nonce=nonce, ttl=ttl)
        logging.info('action=confirmation_request id=%s info=%s' %
                     (id, contact.info))
        if self.func_send:
            # TODO: stop token value from leaking into celery logs
            self.func_send(
                to=id, template=template, token=token, type=contact.type)
        return dict(token=token, id=id, uid=contact.uid), 200

    def confirm(self, token, clear_session=False):
        """Confirm a contact if token is still valid

        Args:
          token (str): the token generated previously
          clear_session (bool): whether to wipe out token after use

        Returns:
          tuple:
            first item is dict with fields as defined in SessionManager.create,
            second item is the http response code
        """

        serializer = URLSafeTimedSerializer(self.token_secret)
        try:
            id, info, nonce = serializer.loads(
                token, salt=self.token_salt,
                max_age=self.token_timeout).split(':')
        except Exception as ex:
            return dict(token=token, message=str(ex)), 403
        if g.session.get(id, nonce, arg='param') != token:
            logging.warning('action=confirm id=%s info=%s '
                            'message=invalid_token' % (id, info))
            return dict(id=id, message='invalid token'), 403
        try:
            contact = g.db.query(self.models.Contact).filter_by(id=id).one()
        except NoResultFound:
            logging.warning('action=confirm id=%s info=%s'
                            'message=Not_found' % (id, info))
            return dict(token=token, id=id, message='not found'), 404
        contact.status = u'active'
        g.db.query(self.models.Person).filter_by(id=contact.uid).update(
            dict(referrer_id=None))
        g.db.commit()

        # Create a login session with permissions to modify a Person/Contact
        logging.info('action=confirm type=%s info=%s' %
                     (contact.type, info))
        auth = 'person'
        if clear_session:
            g.session.delete(id, nonce)
            auth = 'pending'
        return dict(g.session.create(
            contact.uid, [auth], info=info, type=contact.type,
            contact_id=id, name=contact.owner.name)), 200



"""local_func

created 26-mar-2020 by richb@instantlinux.net
"""
from flask import g
from flask_babel import _
import logging
from passlib.hash import sha256_crypt
from sqlalchemy.exc import OperationalError
from sqlalchemy.orm.exc import NoResultFound
from sqlalchemy_utils.types.encrypted.padding import InvalidPaddingError

from .. import ServiceConfig
from ..database import db_abort


def login(username, password):
    """Login using credentials stored in local database.  If
    successful, this passes a sqlalchemy account record back for
    further processing by SessionAuth.login_accepted.

    Args:
      username (str): the username or email identity
      password (str): password

    Returns:
      tuple: 3 items
        dict with error message
        http status (200 if OK)
        account (sqlalchemy query object)
    """
    models = ServiceConfig().models
    try:
        if '@' in username:
            account = g.db.query(models.Account).join(
                models.Person).filter(
                    models.Person.identity == username,
                    models.Account.status == 'active').one()
            username = account.name
        else:
            account = g.db.query(models.Account).filter_by(
                name=username, status='active').one()
    except InvalidPaddingError:
        logging.error('action=login message="invalid DB_AES_SECRET"')
        return dict(message=_(u'DB operational error')), 503, None
    except (NoResultFound, TypeError):
        return dict(username=username, message=_(u'not valid')), 403, None
    except OperationalError as ex:
        content, status = db_abort(str(ex), action='login')
        return content, status, None

    if account.password == '':
        logging.error("username=%s, message='no password'" % username)
        return dict(username=username, message=_(u'no password')), 403, None
    elif not password or not sha256_crypt.verify(password, account.password):
        logging.info(dict(action='login', method='local',
                          username=username, credential='invalid'))
        g.db.commit()
        return dict(username=username, message=_(u'not valid')), 403, None

    return {}, 200, account



"""geocode.py

Geocode functions

created 7-mar-2020 by richb@instantlinux.net
"""

# from geoalchemy2 import functions
import geocoder
import logging

from .access import AccessControl
from .metrics import Metrics


def lookup(address=None, neighborhood=None, city=None, state=None,
           country=None):
    """Geo-code an address and get its neighbhorhood name
    To make this compatible with any database, coordinates are stored
    as integers with 7-digit decimal precision to fit in 32 bits

    Args:
      address (str): street address
      neighborhood (str): neighborhood name
      city (str): city
      state (str): state
      country (str): 2-letter country abbreviation
    Returns:
      tuple: lat, long, neighborhood
    """

    location = '%s,%s,%s' % (city, state, country)
    if address or neighborhood:
        location = '%s,%s' % (address or neighborhood, location)
    try:
        geo = geocoder.mapquest(location)
    except Exception as ex:
        if 'Provide API Key' in str(ex):
            logging.error('geocode: no MapQuest key set')
        else:
            logging.warning('action=geo_lookup error=%s' % str(ex))
        return (None, None, '')

    # POINT column type only supported in PostgreSQL / sqlite
    # geo = 'POINT(%f %f)' % (geo.lng, geo.lat)
    geolat = int(geo.lat * 1.0e7)
    geolong = int(geo.lng * 1.0e7)
    if address and not neighborhood:
        # TODO neighborhood-lookup support
        neighborhood = 'A neighborhood%s' % (' in %s' % city if city else '')
    logging.info('geocode: lat=%f long=%f neighborhood="%s"' % (
        geo.lat, geo.lng, neighborhood))
    Metrics().store('geo_lookup_total')
    return (geolat, geolong, neighborhood)


def with_privacy(record, access):
    """Convert a database record's geo values for API response
    Input is a value from model as_dict.  Evaluates uid,
    geolat, geolong, address fields. If the record is private,
    last 4 digits are rounded off.

    Args:
      record (dict): dict of key/value pairs from database
      access (str): access required for full location disclosure
    Returns:
      dict: record geolat/geolong values removed and geo tuple added
    """

    if access == 'r' or record['uid'] == AccessControl().uid:
        if record.get('geolat') and record.get('geolong'):
            record['geo'] = [record['geolong'] / 1.0e7,
                             record['geolat'] / 1.0e7]
    else:
        if record.get('geolat') and record.get('geolong'):
            record['geo'] = [round(record['geolong'] / 1.0e7, 4),
                             round(record['geolat'] / 1.0e7, 4)]
            if 'address' in record:
                record['address'] = ''
    record.pop('geolong', None)
    record.pop('geolat', None)
    return record



"""ldap_func

created 4-may-2021 by richb@instantlinux.net
"""
from flask import g
from flask_babel import _
from ldap3 import Connection as LDAPConnection, Server as LDAPServer, \
    ServerPool as LDAPServerPool
from ldap3.core.exceptions import LDAPBindError
import logging
from sqlalchemy.orm.exc import NoResultFound

from .. import state
from ..database import db_abort
from ..service_config import ServiceConfig
from ..utils import identity_normalize, utcnow


def login(username, password):
    """Login using LDAP credentials.  If successful, this passes a
    sqlalchemy account record back for further processing by
    SessionAuth.login_accepted. The email identity is either looked
    up from the LDAP attributes (by default, the attr_identity is
    'mail'), or composed as <attr_name>@<email_domain>.

    Args:
      username (str): the username or email identity
      password (str): password

    Returns:
      tuple: 3 items
        dict with error message
        http status (200 if OK)
        account (sqlalchemy query object)
    """
    from ..session_auth import SessionAuth

    method = 'ldap'
    config = ServiceConfig().config
    models = state.models
    logmsg = dict(action='login', method=method, username=username)
    bind_dn = username
    if config.LDAP_PARAMS['domain'] and '\\' not in username:
        bind_dn = config.LDAP_PARAMS['domain'] + '\\' + username
    if state.ldap_mock:
        # This is an embarrassing / ugly hack due to absence of support
        # for the bind() call in ldap3 library's MOCK_SYNC feature, along
        # with the difficulty of patching-out the guts of ldap3.Connect
        conn = state.ldap_mock
        if not conn.bind():
            logging.info(dict(credential='invalid', **logmsg))
            return dict(message=_(u'access denied')), 403, None
    else:
        try:
            conn = LDAPConnection(
                state.ldap_serverpool,
                user=bind_dn, password=password, auto_bind=True,
                authentication=config.LDAP_PARAMS['authentication'])
        except LDAPBindError:
            logging.info(dict(credential='invalid', **logmsg))
            return dict(message=_(u'access denied')), 403, None
        except Exception as ex:
            logging.warning(dict(message='LDAP connection failure',
                                 error=str(ex), **logmsg))
            return dict(message=_(u'access denied')), 403, None
    if state.ldap_conn:
        conn.unbind()
        conn = state.ldap_conn
        conn.bind()
    if not conn.search(
            search_base=config.LDAP_PARAMS['search_base'],
            search_filter=config.LDAP_PARAMS['search_filter'].format(
                user=username),
            search_scope='SUBTREE',
            attributes='*',
            get_operational_attributes=True):
        logging.info(dict(message='search failed', **logmsg))
        return dict(message=_(u'access denied')), 403, None
    user = {key: ''.join(val)
            for (key, val) in conn.response[0]['attributes'].items()
            if type(val) is str or type(val[0]) is str}
    account_name = user.get(config.LDAP_PARAMS['attr_name'])
    if config.LDAP_PARAMS['attr_identity']:
        identity = identity_normalize(user.get(
            config.LDAP_PARAMS['attr_identity']))
    else:
        identity = '%s@%s' % (account_name, config.LDAP_PARAMS['email_domain'])
    logmsg['identity'] = identity
    if not identity or '@' not in identity:
        logging.warning(dict(message='identity missing', usermeta=user,
                             **logmsg))
        return dict(message=_(u'access denied')), 403, None
    try:
        account = g.db.query(models.Account).join(models.Person).filter(
            models.Person.identity == identity,
            models.Account.name == account_name,
            models.Account.status == 'active').one()
    except NoResultFound:
        account = None
    except Exception as ex:
        content, status = db_abort(str(ex), **logmsg)
        return content, status, None
    if not account:
        name = user.get('displayName') or user.get('cn')
        username = user.get(config.LDAP_PARAMS['attr_name'])
        ses = SessionAuth()
        content, status = ses.register(identity, username, name, template=None)
        if status != 200:
            return content, status, None
        content, status = ses.account_add(username, uid=content['uid'])
        if status != 201:
            return content, status, None
        account = g.db.query(models.Account).filter_by(
            id=content['id']).one()
    return {}, 200, account


def ldap_init(ldap_conn=None, ldap_mock=None):
    """Initialize LDAP server pool

    Args:
      ldap_conn (obj): an open connection (for testing)
      ldap_mock (obj): a mock (also for testing)
    """
    start = utcnow().timestamp()
    config = ServiceConfig().config

    if config.LDAP_PARAMS['servers']:
        state.ldap_serverpool = LDAPServerPool(
            None,
            pool_strategy=config.LDAP_PARAMS['pool_strategy'],
            active=config.LDAP_PARAMS['pool_active'],
            exhaust=config.LDAP_PARAMS['pool_strategy'])
        for name in config.LDAP_PARAMS['servers']:
            state.ldap_serverpool.add(LDAPServer(
                '%s://%s' % (config.LDAP_PARAMS['scheme'], name),
                port=config.LDAP_PARAMS['port'],
                allowed_referral_hosts=('*', True)))
        try:
            if config.LDAP_PARAMS['bind_dn']:
                state.ldap_conn = LDAPConnection(
                    state.ldap_serverpool, user=config.LDAP_PARAMS['bind_dn'],
                    password=config.LDAP_BIND_PASSWORD,
                    auto_bind=True, read_only=True,
                    authentication=config.LDAP_PARAMS['authentication'])
            logging.info(dict(
                action='initialize', provider='ldap',
                servers=','.join(config.LDAP_PARAMS['servers']),
                duration='%.3f' % (utcnow().timestamp() - start)))
        except Exception as ex:
            logging.error(dict(action='ldap_init', error=str(ex),
                               message='LDAP connection failure'))
    if ldap_mock:
        state.ldap_mock = ldap_mock
        state.ldap_conn = ldap_conn



"""session_auth

created 26-mar-2020 by richb@instantlinux.net
"""
from flask import g, redirect, request
from flask_babel import _
import jwt
import logging
import os
from sqlalchemy.exc import DataError, IntegrityError
from sqlalchemy.orm.exc import NoResultFound
import time

from . import state
from .access import AccessControl
from .account_settings import AccountSettings
from .const import Constants
from .database import db_abort
from .messaging.confirmation import Confirmation
from .metrics import Metrics
from .ratelimit import RateLimit
from .service_config import ServiceConfig
from .service_registry import ServiceRegistry
from .utils import gen_id, identity_normalize, utcnow

OC = {}


class SessionAuth(object):
    """Session Authorization

    Functions for login, password and role authorization

    Args:
      roles_from (obj): model for which to look up authorizations
    """

    def __init__(self, roles_from=None):
        config = self.config = ServiceConfig().config
        self.jwt_secret = config.JWT_SECRET
        self.login_session_limit = config.LOGIN_SESSION_LIMIT
        self.login_admin_limit = config.LOGIN_ADMIN_LIMIT
        self.func_send = state.func_send
        self.models = state.models
        self.oauth = state.oauth['init']
        self.roles_from = roles_from
        self.redis_conn = state.redis_conn

    def account_login(self, username, password, method=None, otp=None,
                      nonce=None):
        """Log in using local or OAuth2 credentials

        Args:
          username (str): account name or email
          password (str): credential
          method (str): local, ldap, or google / facebook / twitter etc
          otp (str): one-time or backup password
          nonce (str): a nonce check value (for OAuth2: optional)

        Returns:
          dict:
            Fields include jwt_token (contains uid / account ID),
            ID of entry in settings database, and a sub-dictionary
            with mapping of endpoints registered to microservices
        """
        from .auth import ldap_func, local_func, oauth2_func, totp_func

        logmsg = dict(action='login', username=username, method=method)
        acc = AccessControl()
        content, status, headers = _(u'access denied'), 403, []
        if RateLimit(enable=True,
                     interval=self.config.LOGIN_LOCKOUT_INTERVAL).call(
                         limit=self.config.LOGIN_ATTEMPTS_MAX,
                         service='login.attempt', uid=username):
            time.sleep(5)
            msg = _(u'locked out')
            logging.warning(dict(message=msg, **logmsg))
            return dict(username=username, message=msg), 403
        if acc.auth and ('pendingtotp' in acc.auth or (otp and acc.apikey_id)):
            # TODO make this pendingtotp finite-state logic more robust
            if otp:
                content, status, headers, account = totp_func.login(
                    username, otp, redis_conn=self.redis_conn)
                if status == 200:
                    content, status, headers = self.login_accepted(
                        username, account, acc.auth_method, headers=headers)
            else:
                logging.info(dict(error='otp token omitted', **logmsg))
        elif method in self.config.AUTH_PARAMS.keys():
            content, status = oauth2_func.login(self.oauth, method,
                                                nonce=nonce)
        elif not method or method in self.config.AUTH_METHODS:
            items = method if method else self.config.AUTH_METHODS
            for item in items:
                if item == 'local':
                    content, status, account = local_func.login(username,
                                                                password)
                elif item == 'ldap':
                    content, status, account = ldap_func.login(username,
                                                               password)
                if status == 200:
                    method = item
                    break
            if status == 200:
                content, status, headers = self.login_accepted(
                    username, account, method)
        else:
            msg = _(u'unsupported login method')
            logging.error(dict(message=msg, **logmsg))
            return dict(message=msg), 500
        if status in (200, 201):
            RateLimit().reset(service='login.attempt', uid=username)
        else:
            Metrics().store('logins_fail_total', labels=['method=%s' % method])
        return content, status, headers

    def login_accepted(self, username, account, method, headers=None):
        """Login accepted from provider: create a session

        Args:
          username (str): the account's unique username
          account (obj): account object in database
          method (str): method, e.g. local or google
          headers (dict): additional headers, such as Set-Cookie
        """
        # connexion doesn't support cookie auth. probably just as well,
        #  this forced me to implement a JWT design with nonce token
        # https://github.com/zalando/connexion/issues/791
        #  TODO come up with a standard approach, this is still janky
        #  and does not do signature validation (let alone PKI)

        # TODO research concerns raised in:
        # https://paragonie.com/blog/2017/03/jwt-json-web-tokens-is-bad-
        #  standard-that-everyone-should-avoid
        acc = AccessControl()
        logmsg = dict(action='login', username=username, uid=account.uid,
                      method=method, account_id=account.id)
        duration = (self.login_admin_limit if account.is_admin
                    else self.login_session_limit)
        if (account.password_must_change and method == 'local'):
            roles = ['person', 'pwchange']
            duration = 300
        elif not account.totp and self.config.LOGIN_MFA_REQUIRED and (
                method in ('local', 'ldap') or self.config.LOGIN_MFA_EXTERNAL):
            roles = ['mfarequired']
            duration = 300
        elif not self.totp_bypass(account.uid) and account.totp and (
                method in ('local', 'ldap') or
                self.config.LOGIN_MFA_EXTERNAL) and not (
                    acc.auth and 'pendingtotp' in acc.auth):
            roles = ['pendingtotp']
            logmsg['action'] = 'totp_challenge'
            duration = 120
        # TODO parameterize model name contact
        elif (g.db.query(self.models.Contact).filter(
                self.models.Contact.info == account.owner.identity,
                self.models.Contact.type == 'email'
                ).one().status in ('active', 'unconfirmed')):
            roles = ['admin', 'user'] if account.is_admin else ['user']
            if self.roles_from:
                roles += self.get_roles(account.uid,
                                        member_model=self.roles_from)
            try:
                account.last_login = utcnow()
                g.db.commit()
            except Exception as ex:
                logging.error(dict(message='DB operational error',
                                   error=str(ex), **logmsg))
        else:
            roles = []
        if acc.apikey_id:
            logmsg['apikey'] = acc.apikey_id
        logging.info(logmsg)
        ses = g.session.create(account.uid, roles, acc=account.id,
                               identity=account.owner.identity,
                               method=method, ttl=duration)
        if ses:
            retval = dict(
                jwt_token=jwt.encode(
                    ses, self.jwt_secret, algorithm='HS256'),
                resources=ServiceRegistry().find()['url_map'],
                settings_id=account.settings_id, username=account.name)
            if hasattr(account.settings, 'default_storage_id'):
                retval['storage_id'] = account.settings.default_storage_id
            if 'pendingtotp' not in roles:
                Metrics().store('logins_success_total', labels=['method=%s' %
                                                                method])
            if method in ('local', 'ldap'):
                return retval, 201, headers
            else:
                return redirect('%s%s?token=%s' % (
                    account.settings.url,
                    self.config.AUTH_LANDING_PAGE, retval['jwt_token']))
        else:
            return db_abort(_(u'login session trouble'), action='login')

    def totp_bypass(self, uid):
        """Check for bypass cookie

        Args:
          uid (str): User ID
        Returns:
          bool: valid bypass found
        """
        if request.cookies and (self.config.LOGIN_MFA_COOKIE_NAME
                                in request.cookies):
            if g.session.get(
                    uid, request.cookies[self.config.LOGIN_MFA_COOKIE_NAME]):
                return True
        return False

    def register(self, identity, username, name, template='confirm_new',
                 picture=None):
        """Register a new account: create related records in database
        and send confirmation token to new user

        TODO caller still has to invoke account_add function to
        generate record in accounts table

        Args:
          identity (str): account's primary identity, usually an email
          username (str): account's username
          name (str): name
          picture (url): URL of an avatar / photo
          template (str): template for message (confirming new user)
        Returns:
          tuple: the Confirmation.request dict and http response
        """
        if not username or not identity or not name:
            return dict(message=_(u'all fields required')), 400
        identity = identity_normalize(identity)
        logmsg = dict(action='register', identity=identity,
                      username=username)
        try:
            g.db.query(self.models.Account).filter_by(name=username).one()
            msg = _(u'that username is already registered')
            logging.warning(dict(message=msg, **logmsg))
            return dict(message=msg), 405
        except NoResultFound:
            pass
        uid = None
        try:
            existing = g.db.query(self.models.Person).filter_by(
                identity=identity).one()
            uid = existing.id
            g.db.query(self.models.Account).filter_by(uid=uid).one()
            msg = _(u'that email is already registered, use forgot-password')
            logging.warning(dict(message=msg, **logmsg))
            return dict(message=msg), 405
        except NoResultFound:
            pass
        if uid:
            try:
                cid = g.db.query(self.models.Contact).filter_by(
                    info=identity, type='email').one().id
            except Exception as ex:
                return db_abort(str(ex), msg=_(u'contact lookup'), **logmsg)
        else:
            person = self.models.Person(
                id=gen_id(prefix='u-'), name=name, identity=identity,
                status='active')
            uid = person.id
            cid = gen_id()
            try:
                g.db.add(person)
                g.db.add(self.models.Contact(id=cid, uid=uid, type='email',
                                             info=identity))
                g.db.commit()
            except Exception as ex:
                return db_abort(str(ex), rollback=True,
                                msg=_(u'contact activation problem'), **logmsg)
            logging.info(dict(message=_(u'person added'), uid=uid, **logmsg))
        # If browser language does not match global default language, add
        # a profile setting
        lang = request.accept_languages.best_match(self.config.LANGUAGES)
        try:
            default_lang = g.db.query(self.models.Settings).filter_by(
                name='global').one().lang
        except Exception:
            default_lang = 'en'
        if lang and lang != default_lang:
            try:
                g.db.add(self.models.Profile(id=gen_id(), uid=uid, item='lang',
                                             value=lang, status='active'))
                g.db.commit()
            except IntegrityError:
                # Dup entry from previous attempt
                g.db.rollback()
        if picture:
            try:
                g.db.add(self.models.Profile(
                    id=gen_id(), uid=uid, item='picture', value=picture,
                    status='active'))
                g.db.commit()
            except (DataError, IntegrityError):
                # Dup entry from previous attempt or value too long
                g.db.rollback()
        if template:
            return Confirmation().request(cid, template=template)
        else:
            try:
                g.db.query(self.models.Contact).filter_by(
                    id=cid).one().status = 'active'
                g.db.commit()
            except Exception as ex:
                return db_abort(str(ex), rollback=True,
                                msg=_(u'contact activation problem'), **logmsg)
            return dict(uid=uid, message='registered'), 200

    def account_add(self, username, uid):
        """Add an account with the given username

        Args:
          username (str): new / unique username
          uid (str): existing user
        """
        logmsg = dict(action='account_add', username=username)
        try:
            settings_id = g.db.query(
                self.models.Settings).filter_by(name='global').one().id
        except Exception as ex:
            return db_abort(str(ex), message=u'failed to read global settings',
                            **logmsg)
        account = self.models.Account(
            id=gen_id(), uid=uid, name=username,
            password_must_change=True, password='',
            settings_id=settings_id, status='active')
        try:
            g.db.add(account)
            g.db.commit()
        except Exception as ex:
            return db_abort(str(ex), rollback=True, **logmsg)
        return dict(id=account.id, uid=uid, username=username), 201

    def auth_params(self):
        """Get authorization info"""
        account_id = AccessControl().account_id
        settings = AccountSettings(account_id, db_session=g.db)
        try:
            account = g.db.query(self.models.Account).filter_by(
                id=account_id).one()
        except Exception as ex:
            return db_abort(str(ex), action='auth_params')
        retval = dict(
            account_id=account_id, uid=AccessControl().uid,
            resources=ServiceRegistry().find()['url_map'],
            settings_id=settings.get.id, totp=account.totp)
        if hasattr(account.settings, 'default_storage_id'):
            storage_id = settings.get.default_storage_id
            if storage_id:
                retval['storage_id'] = storage_id
        return retval, 200

    def methods(self):
        """Return list of available auth methods"""
        internal_policy = self.config.LOGIN_INTERNAL_POLICY
        if 'local' not in self.config.AUTH_METHODS:
            internal_policy = 'closed'
        ret = []
        for method in self.config.AUTH_METHODS:
            if method == 'oauth2':
                for ext_method in self.config.AUTH_PARAMS.keys():
                    if os.environ.get('%s_CLIENT_ID' % ext_method.upper()):
                        ret.append(ext_method)
            else:
                ret.append(method)
        return dict(
            items=ret, count=len(ret),
            policies=dict(
                login_internal=internal_policy,
                login_external=self.config.LOGIN_EXTERNAL_POLICY)), 200

    def get_roles(self, uid, member_model=None, resource=None, id=None):
        """Get roles that match uid / id for a resource
        Each is in the form <resource>-<id>-<privacy level>

        Args:
          uid (str): User ID
          member_model (str): resource-name of DB model that defines
            membership in resource
          resource (str): the resource that defines privacy (e.g. list)
          id (str): ID of the resource (omit if all are desired)
        Returns:
          list of str: authorized roles
        """
        acc = AccessControl()
        # TODO improve defaulting of member_model and resource from
        #  rbac.yaml
        resource = resource or acc.primary_resource
        if not member_model and len(state.private_res):
            member_model = state.private_res[0].get('resource')
        if not resource or not member_model:
            return []
        column = '%s_id' % resource
        try:
            query = g.db.query(getattr(
                self.models, member_model.capitalize())).filter_by(
                    uid=uid, status='active')
            if id:
                query = query.filter_by(**{column: id})
        except Exception as ex:
            logging.warning('action=get_roles message="%s"' % str(ex))
        roles = []
        for record in query.all():
            if not hasattr(record, resource):
                continue
            # Skip first defined level (always 'public')
            for level in acc.privacy_levels[1:]:
                roles.append('%s-%s-%s' % (resource, getattr(record, column),
                                           level))
                if level == record.authorization:
                    break
        return roles

    def update_auth(self, id, member_model=None, resource=None, force=False):
        """Check current access, update if recently changed

        Args:
          id (str): resource id of parent resource
          resource (str):
            parent resource for which membership should be checked
          force (bool): perform update regardless of logged-in permissions
        """
        acc = AccessControl()
        resource = resource or acc.primary_resource
        current = set(acc.auth if acc.auth else [])
        if force or ('%s-%s-%s' % (resource, id, Constants.AUTH_INVITEE)
                     not in current):
            # TODO handle privilege downgrade member/host->invitee
            current |= set(self.get_roles(acc.uid, member_model=member_model,
                                          resource=resource, id=id))
            creds = request.authorization
            g.session.update(creds.username, creds.password, 'auth',
                             ':'.join(current))



"""oauth2_func

created 29-mar-2021 by richb@instantlinux.net
"""
from authlib.integrations.flask_client import OAuth
from flask_babel import _
import logging
import os

from .. import ServiceConfig, state


def login(oauth_reg, method, nonce=None):
    """Initiate OAuth2 login

    Args:
      oauth_reg (obj): pre-registered oauth object (from initialize.py)
      method (str): provider name
      nonce (str): a nonce check value
    """
    logmsg = dict(action='account_login', method=method)
    try:
        client = oauth_reg.create_client(method)
    except RuntimeError as ex:
        msg = _(u'login client missing')
        logging.error(dict(message=msg, error=str(ex), **logmsg))
        return dict(message=msg), 500
    config = ServiceConfig().config
    msg = _(u'openid client failure')
    retval = client.authorize_redirect(
        '%s%s/%s/%s' % (config.PUBLIC_URL, config.BASE_URL,
                        'auth_callback', method), nonce=nonce)
    if retval.status_code == 302:
        return dict(location=retval.location), 200
    logging.error(dict(status_code=retval.status_code, **logmsg))
    return dict(message=msg), 405


def oauth2_init(app):
    """Register the OAuth2 providers specified in AUTH_PARAMS config.
    The client-id and client-secret params must be specified as
    environment variables, in the form GOOGLE_CLIENT_ID.

    Args:
      app(obj): the flask object
    """
    state.oauth['init'] = OAuth(app)
    config = ServiceConfig().config
    for provider in config.AUTH_PARAMS.keys():
        client_id = os.environ.get('%s_CLIENT_ID' % provider.upper())
        client_secret = os.environ.get('%s_CLIENT_SECRET' % provider.upper())
        if client_id:
            state.oauth['init'].register(name=provider, client_id=client_id,
                                         client_secret=client_secret,
                                         **config.AUTH_PARAMS[provider])
            logging.info(dict(action='initialize', provider=provider))



from flask import g, request
import logging

from apicrud import AccessControl, BasicCRUD, state, utils


class PersonController(BasicCRUD):
    def __init__(self):
        super().__init__(resource='person')

    @staticmethod
    def create(body):
        """create the first contact record after adding the person

        params:
          body: REST payload
        """
        body['referrer_id'] = AccessControl().uid
        retval = super(PersonController, PersonController).create(
            body, id_prefix='u-')
        if retval[1] != 201:
            return retval
        self = state.controllers[request.url_rule.rule.split('/')[3]]
        record = state.models.Contact(
            id=utils.gen_id(), uid=retval[0]['id'],
            type='email', info=body['identity'], status='unconfirmed')
        g.db.add(record)
        try:
            g.db.commit()
        except Exception as ex:
            logging.warn(dict(action='create', resource=self.resource,
                              message=str(ex)))
            return dict(message=str(ex), data=str(body)), 405
        return dict(id=body['id']), 201
